# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP RIB in and out policy
"""


class Rules(dict):
    """Route Policy Rules"""

    def __init__(self, rules=None, default_rule=None):
        """Initialize the Route Policy Rules"""

        super(Rules, self).__init__(rules or {})
        self.default_rule = default_rule

    @classmethod
    def from_dict(cls, rules_dict, default_rule=None):
        pass

    @classmethod
    def from_database(cls):
        pass



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class AtomicAggregate(Attribute):
    """
    ATOMIC_AGGREGATE is a well-known discretionary attribute of length 0.
    """

    ID = AttributeID.ATOMIC_AGGREGATE
    FLAG = AttributeFlag.TRANSITIVE

    @classmethod
    def parse(cls, value):

        """
        parse bgp ATOMIC_AGGREGATE attribute
        :param value:
        """
        if not value:
            return value
        else:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_OPTIONAL_ATTR,
                data=value)

    @classmethod
    def construct(cls, value):
        """construct a ATOMIC_AGGREGATE path attribute
        :param value:
        """

        if value:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_OPTIONAL_ATTR,
                data='')
        else:
            value = 0
        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
            + struct.pack('!B', value)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri.ipv4_flowspec import IPv4FlowSpec
from yabgp.message.attribute.nlri.ipv4_flowspec import operator_str_dict


class TestIPv4FlowSpec(unittest.TestCase):

    def test_parse(self):
        pass
        raw_hex = '\x01\x18\x02\x02\x02\x02\x10\x03\x03\x03\x01\x00\x01\x2f\x01\x58' \
                  '\x01\x01\x01\x02\x01\x59\x81\x67\x05\x11\x1f\x92\x91\x1f\x93\x06\x01' \
                  '\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93\x07\x01\x02\x01' \
                  '\x03\x01\x05\x81\x06\x08\x81\x02\x09\x81\x28\x0a\x01\xfe\x03\xfe\xd5' \
                  '\x01\x2c\x0b\x01\x28\x81\x30\x0c\x81\x01'
        self.assertEqual(
            [{1: '2.2.2.0/24'}, {2: '3.3.0.0/16'}, {3: '=0 =47 =88 =1 =2 =89 =103'}, {5: '=8082 =8083'},
             {6: '=80 =8080 =8081 =8082 =8083'}, {7: '=2 =3 =5 =6'}, {8: '=2'}, {9: '=40'}, {10: '=254 >=254&<=300'},
             {11: '=40 =48'}, {12: '=1'}], IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_prefix(self):
        raw_hex = '\x01\x18\x6e\x01\x01'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_DST_PFIX: '110.1.1.0/24'}], IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_packet_length(self):
        raw_hex = '\x0a\x01\xfe\x03\xfe\xd5\x01\x2c'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_PCK_LEN: '=254 >=254&<=300'}], IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_ip_protocol(self):
        raw_hex = '\x03\x01\x00\x01\x2f\x01\x58\x01\x01\x01\x02\x01\x59\x81\x67'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_IP_PROTO: '=0 =47 =88 =1 =2 =89 =103'}], IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_des_port(self):
        raw_hex = '\x05\x01\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_DST_PORT: '=80 =8080 =8081 =8082 =8083'}],
                         IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_src_port(self):
        raw_hex = '\x06\x01\x50\x11\x1f\x90\x11\x1f\x91\x11\x1f\x92\x91\x1f\x93'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_SRC_PORT: '=80 =8080 =8081 =8082 =8083'}],
                         IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_icmp_type(self):
        raw_hex = '\x07\x01\x02\x01\x03\x01\x05\x81\x06'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_ICMP_TP: '=2 =3 =5 =6'}], IPv4FlowSpec.parse(raw_hex))

    def test_parse_nlri_icmp_code(self):
        raw_hex = '\x08\x81\x02'
        self.assertEqual([{bgp_cons.BGPNLRI_FSPEC_ICMP_CD: '=2'}], IPv4FlowSpec.parse(raw_hex))

    def test_operator_str_dict(self):
        operator_str1 = '=123'
        operator_str2 = '<=123'
        operator_str3 = '<=1024 >=2048'
        operator_str4 = '<=1024 >=2048&<=3789'
        print operator_str_dict(operator_str1)
        print operator_str_dict(operator_str2)
        print operator_str_dict(operator_str3)
        print operator_str_dict(operator_str4)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test AS PATH attribute """

import unittest

from yabgp.common.exception import UpdateMessageError
from yabgp.common.constants import ERR_MSG_UPDATE_MALFORMED_ASPATH
from yabgp.message.attribute.aspath import ASPath


class TestASPATH(unittest.TestCase):
    def test_parse(self):

        as_path = ASPath.parse(value=b'')
        self.assertEqual(as_path, [])

        # 2bytes ASN
        as_path = ASPath.parse(value=b'\x02\x04\x0c\xb9y3\x88 S\xd9')
        self.assertEqual(as_path, [(2, [3257, 31027, 34848, 21465])])

        # 4bytes ASN
        as_path = ASPath.parse(value=b'\x02\x04\x00\x00\x0c\xb9\x00\x00y3\x00\x00\x88 \x00\x00S\xd9',
                               asn4=True)
        self.assertEqual(as_path, [(2, [3257, 31027, 34848, 21465])])

        # MALFORMED_ASPATH
        try:
            ASPath.parse(value=b'\x02\x04\x00\x00\x0c\xb9\x00\x00y3\x00\x00\x88 \x00\x00S\xd9')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_MALFORMED_ASPATH)
        try:
            ASPath.parse(value=b'\x05\x04\x0c\xb9y3\x88 S\xd9')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_MALFORMED_ASPATH)

    def test_parse_as_set_as_federate(self):
        as_path = ASPath.parse(value=b'\x04\x02\x03\xe9\x03\xea\x03\x02\x03\xeb\x03\xec')
        self.assertEqual(as_path, [(4, [1001, 1002]), (3, [1003, 1004])])

    def test_construct(self):

        # 2 bytes ASN
        as_path = ASPath.construct(asn4=False, value=[(2, [3257, 31027, 34848, 21465])])
        self.assertEqual(as_path, b'@\x02\n\x02\x04\x0c\xb9y3\x88 S\xd9')

        # 4 bytes ASN
        as_path = ASPath.construct(asn4=True, value=[(2, [3257, 31027, 34848, 21465])])
        self.assertEqual(as_path, b'@\x02\x12\x02\x04\x00\x00\x0c\xb9\x00\x00y3\x00\x00\x88 \x00\x00S\xd9')

    def test_construct_as_set(self):
        as_path = ASPath.construct(asn4=False, value=[(2, [1001, 1002]), (1, [1003, 1004])])
        self.assertEqual(as_path, b'@\x02\x0c\x02\x02\x03\xe9\x03\xea\x01\x02\x03\xeb\x03\xec')

    def test_construct_as_set_as_federate(self):
        as_path = ASPath.construct(asn4=False, value=[(4, [1001, 1002]), (3, [1003, 1004])])
        self.assertEqual(as_path, b'@\x02\x0c\x04\x02\x03\xe9\x03\xea\x03\x02\x03\xeb\x03\xec')


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""IPv4 Flowspec NLRI
"""
import struct
import math
import binascii

import netaddr

from yabgp.common import constants as bgp_cons


class IPv4FlowSpec(object):

    @classmethod
    def parse(cls, value):
        """
        parse IPv4 flowspec NLRI
        :param value:
        :return:
        """
        # +------------------------------+
        # |    length (0xnn or 0xfn nn)  |
        # +------------------------------+
        # |    NLRI value  (variable)    |
        # +------------------------------+
        nlri_list = []
        while value:
            offset = 0
            flowspec_type = struct.unpack('!B', value[0])[0]
            offset += 1
            # decode all kinds of flow spec
            if flowspec_type in [bgp_cons.BGPNLRI_FSPEC_DST_PFIX, bgp_cons.BGPNLRI_FSPEC_SRC_PFIX]:
                prefix, offset_tmp = cls.parse_prefix(value[offset:])
                offset += offset_tmp
                nlri_list.append({flowspec_type: prefix})
                value = value[offset:]
            elif flowspec_type in [bgp_cons.BGPNLRI_FSPEC_IP_PROTO, bgp_cons.BGPNLRI_FSPEC_DST_PORT,
                                   bgp_cons.BGPNLRI_FSPEC_SRC_PORT, bgp_cons.BGPNLRI_FSPEC_ICMP_TP,
                                   bgp_cons.BGPNLRI_FSPEC_ICMP_CD,
                                   bgp_cons.BGPNLRI_FSPEC_PCK_LEN]:
                operator_list, offset = cls.parse_operators(value[offset:])
                nlri_list.append({flowspec_type: operator_dict_to_str(operator_list)})
                value = value[offset:]
            else:
                operator_list, offset = cls.parse_operators(value[offset:])
                nlri_list.append({flowspec_type: operator_dict_to_str(operator_list)})
                value = value[offset:]
        return nlri_list

    @classmethod
    def construct(cls, value):
        return cls.construct_nlri(value)

    @classmethod
    def construct_nlri(cls, data):
        """ Construct NLRI """
        nlri_data = ''
        for nlri in data:
            nlri_tmp = ''
            # there may have many filters in each nlri
            for f_type in nlri:
                if int(f_type) in [bgp_cons.BGPNLRI_FSPEC_DST_PFIX, bgp_cons.BGPNLRI_FSPEC_SRC_PFIX]:
                    nlri_tmp += struct.pack('!B', int(f_type)) + cls.construct_prefix(nlri[f_type])
                elif int(f_type) in [bgp_cons.BGPNLRI_FSPEC_IP_PROTO, bgp_cons.BGPNLRI_FSPEC_DST_PORT,
                                     bgp_cons.BGPNLRI_FSPEC_SRC_PORT, bgp_cons.BGPNLRI_FSPEC_ICMP_TP,
                                     bgp_cons.BGPNLRI_FSPEC_ICMP_CD, bgp_cons.BGPNLRI_FSPEC_PCK_LEN]:
                    # translate from expression to binary
                    # expr = '>8080&<8088|=3128'
                    if nlri[f_type] > 255:
                        nlri_tmp += struct.pack('!B', int(f_type)) + '\x91' + struct.pack('!H', nlri[f_type])
                    else:
                        nlri_tmp += struct.pack('!B', int(f_type)) + '\x81' + struct.pack('!B', nlri[f_type])
                    # TODO(penxiao) other flow type
            # nlri_data += struct.pack('!H', len(nlri_tmp)) + nlri_data
            if len(nlri_tmp) >= 240:
                nlri_hex = struct.pack('!H', len(nlri_tmp)) + nlri_tmp
                nlri_data += nlri_hex
            elif nlri_tmp:
                nlri_data += struct.pack('!B', len(nlri_tmp)) + nlri_tmp
        return nlri_data
        #     return struc

    @staticmethod
    def parse_prefix(data):
        """
        Prefixes are encoded as in BGP UPDATE messages, a length in bits is followed by
        enough octets to contain the prefix information.

        Encoding: <prefix-length (1 octet), prefix>
        """
        prefix_len = struct.unpack('!B', data[0])[0]
        octet_len = int(math.ceil(prefix_len / 8.0))
        tmp = data[1:octet_len + 1]
        prefix_data = [ord(i) for i in tmp]
        prefix_data = prefix_data + list(str(0)) * 4
        prefix = "%s.%s.%s.%s" % (tuple(prefix_data[0:4])) + '/' + str(prefix_len)
        return prefix, octet_len + 1

    @classmethod
    def construct_prefix(cls, prefix):
        """
        construct a prefix string from '1.1.1.0/24' to '\x18\x01\x01\x01'
        """
        ip, masklen = prefix.split('/')
        ip_hex = netaddr.IPAddress(ip).packed
        masklen = int(masklen)
        if 16 < masklen <= 24:
            ip_hex = ip_hex[0:3]
        elif 8 < masklen <= 16:
            ip_hex = ip_hex[0:2]
        elif 0 < masklen <= 8:
            ip_hex = ip_hex[0:1]
        elif masklen == 0:
            ip_hex = ''
        return struct.pack('!B', masklen) + ip_hex

    @staticmethod
    def parse_operators(data):
        offset = 0
        parse_operator_list = []
        while data:
            operator = parse_operator(struct.unpack('!B', data[0])[0])
            offset += 1
            operator_value = int(binascii.b2a_hex(data[1:1 + operator['LEN']]), 16)
            offset += operator['LEN']
            parse_operator_list.append([operator, operator_value])
            # the end of the list
            data = data[1 + operator['LEN']:]
            if operator['EOL']:
                break
        return parse_operator_list, offset + 1


def parse_operator(data):
    """
    The operator byte is encoded as:
      0    1   2   3   4  5   6   7
    +---+---+---+---+---+---+---+---+
    |EOL|AND|  LEN  |RES|LT |GT |EQ |
    +---+---+---+---+---+---+---+---+
    """
    bit_list = []
    for i in xrange(8):
        bit_list.append((data >> i) & 1)
    bit_list.reverse()
    result = {
        'EOL': bit_list[0],
        'AND': bit_list[1],
        'LEN': 1 << (bit_list[2] * 2 + bit_list[3]),
        'LT': bit_list[5],
        'GT': bit_list[6],
        'EQ': bit_list[7]
    }
    return result


def operator_dict_to_str(data):
    """

    from
    [
        [
            {'AND': 0, 'GT': 0, 'LEN': 1, 'EOL': 0, 'LT': 0, 'EQ': 1},
            254
        ],
        [
            {'AND': 0, 'GT': 1, 'LEN': 1, 'EOL': 0, 'LT': 0, 'EQ': 1},
            254
        ],
        [
            {'AND': 1, 'GT': 0, 'LEN': 2, 'EOL': 1, 'LT': 1, 'EQ': 1},
            300
        ]
    ]
    to
    =254 >=254&<=300
    :param data: dict
    :return: string format
    """
    return_str = ''
    for item in data:
        operator_dict, value = item
        if operator_dict['AND']:
            return_str += '&'
        else:
            if return_str != '':
                return_str += ' '
        if operator_dict['GT']:
            return_str += '>'
        if operator_dict['LT']:
            return_str += '<'
        if operator_dict['EQ']:
            return_str += '='
        return_str += str(value)
    return return_str


def operator_str_dict(data):
    """
    from =254 >=254&<=300 to dict
    :param data:
    :return:
    """
    # operator_list = []
    # pattern1 = re.compile('^(\D+)(\d+)(&)(\D+)(\d+)$')
    # pattern2 = re.compile('^(\D+)(\d+)$')
    #
    # operator = data.strip().split(' ')
    #
    # for item in operator:
    #     first_match = pattern1.match(item)
    #     if first_match:
    #         print first_match.group(1, 2, 3, 4, 5)
    #     second_match = pattern2.match(item)
    #     if second_match:
    #         operator_type, value = second_match.group(1, 2)
    #         if operator_type == '=':
    #             pass



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Attribute MP_REACH_NLRI
"""

import struct

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.common import afn
from yabgp.common import safn
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons
from yabgp.message.attribute.nlri.ipv4_mpls_vpn import IPv4MPLSVPN
from yabgp.message.attribute.nlri.ipv4_flowspec import IPv4FlowSpec


class MpReachNLRI(Attribute):
    """
        MP_REACH_NLRI (type code 14) is used to carry the set of reachable
    destinations together with the next hop information to be used for
    forwarding to these destinations (RFC 4760 page 2).

        MP_REACH_NLRI coding information
        +---------------------------------------------------------+
        | Address Family Identifier (2 octets)                    |
        +---------------------------------------------------------+
        | Subsequent Address Family Identifier (1 octet)          |
        +---------------------------------------------------------+
        | Length of Next Hop Network Address (1 octet)            |
        +---------------------------------------------------------+
        | Network Address of Next Hop (variable)                  |
        +---------------------------------------------------------+
        | Reserved (1 octet)                                      |
        +---------------------------------------------------------+
        | Network Layer Reachability Information (variable)       |
        +---------------------------------------------------------+
    """

    ID = AttributeID.MP_REACH_NLRI
    FLAG = AttributeFlag.OPTIONAL

    @classmethod
    def parse(cls, value):

        try:
            afi, safi, nexthop_length = struct.unpack('!HBB', value[0:4])
            nexthop_data = value[4:4 + nexthop_length]
            nlri_data = value[5 + nexthop_length:]
        except Exception:
            # error when lenght is wrong
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=repr(value))
        if afi == afn.AFNUM_INET:
            if safi == safn.SAFNUM_LAB_VPNUNICAST:
                nlri = IPv4MPLSVPN.parse(nlri_data)

            elif safi == safn.SAFNUM_FSPEC_RULE:
                # if nlri length is greater than 240 bytes, it is encoded over 2 bytes
                if len(nlri_data) >= 240:
                    nlri_data = nlri_data[2:]
                else:
                    nlri_data = nlri_data[1:]
                nlri = IPv4FlowSpec.parse(nlri_data)

            else:
                nlri = repr(nlri_data)
        else:
            nlri = repr(nlri_data)

        return dict(afi_safi=(afi, safi), nexthop=nexthop_data, nlri=nlri)

    @classmethod
    def construct(cls, value):

        """Construct a attribute

        :param value: python dictionary
        {'afi_safi': (1,128),
         'nexthop': {},
         'nlri': []
        """
        afi, safi = value['afi_safi']
        if afi == afn.AFNUM_INET:
            if safi == safn.SAFNUM_LAB_VPNUNICAST:  # MPLS VPN
                pass
            elif safi == safn.SAFNUM_FSPEC_RULE:  # BGP Flow spec
                try:
                    try:
                        nexthop = netaddr.IPAddress(value['nexthop']).packed
                    except netaddr.core.AddrFormatError:
                        nexthop = ''
                    nlri = IPv4FlowSpec.construct(value=value['nlri'])
                    if nlri:
                        attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + \
                            struct.pack('!B', len(nexthop)) + nexthop + '\x00' + nlri
                        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                            + struct.pack('!B', len(attr_value)) + attr_value
                except Exception as e:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes: %s' % e,
                        data=value
                    )
            else:
                raise excep.ConstructAttributeFailed(
                    reason='unsupport this sub address family',
                    data=value)
        else:
            raise excep.ConstructAttributeFailed(
                reason='unsupport this sub address family',
                data=value)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Route Refresh Message"""

import struct


class RouteRefresh(object):

    """
    Route Refresh message
    """

    def __init__(self, afi=None, safi=None, res=0):

        # Message Format: One <AFI, SAFI> encoded as
        # 0       7       15      23     31
        # +-------+-------+-------+-------+
        # |      AFI      | Res.  | SAFI  |
        # +-------+-------+-------+-------+
        """
        :param afi  - Address Family Identifier (16 bit)
               http://www.iana.org/assignments/address-family-numbers/address-family-numbers.xml
        :param safi: . - Reserved (8 bit) field. Should be set to 0 by the
               sender and ignored by the receiver.
        :param res:  - Subsequent Address Family Identifier (8 bit).
               http://www.iana.org/assignments/safi-namespace/safi-namespace.xml
        """
        self.afi = afi
        self.res = res
        self.safi = safi

    def parse(self, msg):

        """ Parse a route refresh message

         :param msg: raw hex message """

        self.afi, self.res, self.safi = struct.unpack("!HBB", msg)
        return self.afi, self.res, self.safi

    @staticmethod
    def construct_header(message, msg_type):

        """Prepends the mandatory header to a constructed BGP message

        :param msg_type: 5 or 128 """
        #    16-octet     2-octet  1-octet
        # ---------------+--------+---------+------+
        #    Maker      | Length |  Type   |  msg |
        # ---------------+--------+---------+------+
        return b'\xff'*16 + struct.pack('!HB',
                                        len(message) + 19,
                                        msg_type) + message

    def construct(self, msg_type):

        """
        construts a BGP Route Refresh message
        Two Types
        type = 5 (new route refresh RFC 2918)
        type = 128 (cisco route refresh before RFC 2918)
        """

        msg = struct.pack('!H', self.afi) + struct.pack('!B', self.res) + struct.pack('!B', self.safi)
        return self.construct_header(msg, msg_type)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Open message"""

import unittest

import netaddr

from yabgp.message.open import Open
from yabgp.common.constants import VERSION


class TestOpen(unittest.TestCase):
    def setUp(self):
        self.open = Open()
        self.maxDiff = None

    def test_parse(self):
        msg_hex = b'\x04\x5b\xa0\x00\xb4\x03\x03\x03\x09\x25\x02\x06\x01\x04\x00\x01\x00\x80' \
                  b'\x02\x06\x01\x04\x00\x01\x00\x01\x02\x02\x80\x00\x02\x02\x02\x00\x02\x03' \
                  b'\x83\x01\x00\x02\x06\x41\x04\x00\x01\x04\x6a'
        open_msg = self.open.parse(msg_hex)
        results = {'bgpID': '3.3.3.9', 'Version': 4, 'holdTime': 180, 'ASN': 66666,
                   'Capabilities': {
                       'cisco_multi_session': True,
                       'cisco_route_refresh': True,
                       'four_bytes_as': True,
                       'afi_safi': [(1, 128), (1, 1)],
                       'route_refresh': True}}
        self.assertEqual(results, open_msg)

    def test_construct(self):
        self.open.version = VERSION
        self.open.asn = 66666
        self.open.hold_time = 180
        self.open.bgp_id = int(netaddr.IPAddress('1.1.1.1'))
        my_capa = {
            'graceful_restart': False,
            'cisco_multi_session': True,
            'cisco_route_refresh': True,
            'four_bytes_as': True,
            'afi_safi': [(1, 128), (1, 1)],
            'route_refresh': True}
        msg_hex = self.open.construct(my_capa)
        hope_hex = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00=\x01\x04[\xa0\x00\xb4\x01' \
                   b'\x01\x01\x01 \x02\x06\x01\x04\x00\x01\x00\x80\x02\x06\x01\x04\x00\x01\x00\x01\x02\x02\x80\x00' \
                   b'\x02\x02\x02\x00\x02\x06A\x04\x00\x01\x04j'
        self.assertEqual(hope_hex, msg_hex)

    def test_construct_add_path(self):
        self.open.version = VERSION
        self.open.asn = 64512
        self.open.hold_time = 180
        self.open.bgp_id = int(netaddr.IPAddress('10.0.0.6'))
        my_capa = {
            'cisco_route_refresh': True,
            'route_refresh': True,
            'add_path': 'ipv4_receive',
            'four_bytes_as': True,
            'afi_safi': [(1, 1)],
            'enhanced_route_refresh': True}
        msg_hex = self.open.construct(my_capa)
        hope_hex = b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x00A\x01\x04' \
                   b'\xfc\x00\x00\xb4\n\x00\x00\x06$\x02\x06\x01\x04\x00\x01\x00\x01\x02\x02\x80' \
                   b'\x00\x02\x02\x02\x00\x02\x06A\x04\x00\x00\xfc\x00\x02\x06E\x04\x00\x01\x01\x01\x02\x02F\x00'

        self.assertEqual(hope_hex, msg_hex)

    def test_parser_add_path(self):
        msg_hex = b'\x04\xfc\x00\x00\xb4\x0a\x00\x00\x06\x24\x02\x06\x01\x04\x00\x01\x00\x01\x02\x02\x80\x00\x02' \
                  b'\x02\x02\x00\x02\x02\x46\x00\x02\x06\x45\x04\x00\x01\x01\x03\x02\x06\x41\x04\x00\x00\xfc\x00'
        open_msg = self.open.parse(msg_hex)
        results = {
            'bgpID': '10.0.0.6',
            'Version': 4,
            'holdTime': 180,
            'ASN': 64512,
            'Capabilities': {
                'cisco_route_refresh': True,
                'route_refresh': True,
                'add_path': 'ipv4_both',
                'four_bytes_as': True,
                'afi_safi': [(1, 1)],
                'enhanced_route_refresh': True}}
        self.assertEqual(results, open_msg)


if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" database config """

from oslo_config import cfg

CONF = cfg.CONF

database_base_options = [
    cfg.StrOpt('connection',
               default='mongodb://127.0.0.1:27017',
               help='Database connection'),
    cfg.StrOpt('dbname',
               default='yabgp',
               help='database name'),
    cfg.BoolOpt('use_replica',
                default=False,
                help='if use replica set')
]

database_replica_options = [

    cfg.StrOpt('replica_name',
               default='rs1',
               help='the replica set name'),
    cfg.IntOpt('read_preference',
               default=3,
               help='the read preference mode for this instance'),
    cfg.IntOpt('write_concern',
               default=-1,
               help='write concern'),
    cfg.IntOpt('write_concern_timeout',
               default=5000,
               help='write concern timeout')
]


def register_options():
    CONF.register_opts(database_base_options, group='database')
    if CONF.database.use_replica:
        CONF.register_opts(database_replica_options, group='database')



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test Nexthop attribute
"""

import unittest

from yabgp.common.exception import UpdateMessageError
from yabgp.common.constants import ERR_MSG_UPDATE_ATTR_LEN
from yabgp.common.constants import ERR_MSG_UPDATE_INVALID_NEXTHOP
from yabgp.message.attribute.nexthop import NextHop


class TestNextHop(unittest.TestCase):

    def test_parse(self):

        next_hop = NextHop.parse(value=b'\x0a\x0a\x0a\x01')
        self.assertEqual(next_hop, '10.10.10.1')
        self.assertEqual(NextHop.parse(value=b'\x00\x00\x00\x00'), '0.0.0.0')
        self.assertEqual(NextHop.parse(value=b'\xff\xff\xff\xff'), '255.255.255.255')

    def test_parse_invalid_length(self):
        # invalid length
        self.assertRaises(UpdateMessageError, NextHop.parse,
                          b'\x0a\x0a\x0a\x01\x01')
        try:
            NextHop.parse(b'\x0a\x0a\x0a\x01\x01')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_ATTR_LEN)

    def test_construct(self):

        next_hop = NextHop.construct(value='10.10.10.1')
        self.assertEqual(next_hop, b'\x40\x03\x04\x0a\x0a\x0a\x01')
        self.assertEqual(NextHop.construct('0.0.0.0'), b'\x40\x03\x04\x00\x00\x00\x00')

    def test_construct_invalid_nexthop(self):
        # invalid next hop
        self.assertRaises(UpdateMessageError, NextHop.construct, '300.1.1')
        try:
            NextHop.construct(value='300.33.33.33')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_INVALID_NEXTHOP)

    def test_construct_with_flags(self):
        nexthop = NextHop.construct(value='10.10.10.10')
        self.assertEqual(nexthop, b'\x40\x03\x04\x0a\x0a\x0a\x0a')
if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test Originator ID attribute
"""

import unittest

from yabgp.common.constants import ERR_MSG_UPDATE_ATTR_LEN
from yabgp.common import exception as excep
from yabgp.message.attribute.originatorid import OriginatorID


class TestOriginatorID(unittest.TestCase):

    def test_parse(self):

        # normal
        originator_id = OriginatorID.parse(value=b'\xc0\xa8\x01\x01')
        self.assertEqual('192.168.1.1', originator_id)

        # invalid attribute length
        self.assertRaises(excep.UpdateMessageError, OriginatorID.parse,
                          b'\xc0\xa8\x01\x01\x01')
        try:
            OriginatorID.parse(value=b'\xc0\xa8\x01\x01\x01')
        except excep.UpdateMessageError as e:
            self.assertEqual(ERR_MSG_UPDATE_ATTR_LEN, e.sub_error)

    def test_construct(self):

        originator_id = OriginatorID.construct(value='192.168.1.1')
        self.assertEqual(b'\x80\x09\x04\xc0\xa8\x01\x01', originator_id)

        # invalid value
        self.assertRaises(excep.UpdateMessageError, OriginatorID.construct,
                          b'\xc0\xa8\x01\x01\x01')
        try:
            OriginatorID.construct(value=b'\xc0\xa8\x01\x01\x01')
        except excep.UpdateMessageError as e:
            self.assertEqual(ERR_MSG_UPDATE_ATTR_LEN, e.sub_error)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Address family numbers, from
http://www.iana.org/assignments/address-family-numbers
"""

AFNUM_RESERVED = 0  # Reserved
AFNUM_INET = 1  # IP (IP version 4)
AFNUM_INET6 = 2  # IP6 (IP version 6)
AFNUM_NSAP = 3  # NSAP
AFNUM_HDLC = 4  # HDLC (8-bit multidrop)
AFNUM_BBN1822 = 5  # BBN 1822
AFNUM_802 = 6  # 802 (includes all 802 media plus Ethernet "canonical format")
AFNUM_E163 = 7  # E.163
AFNUM_E164 = 8  # E.164 (SMDS, Frame Relay, ATM)
AFNUM_F69 = 9  # F.69 (Telex)
AFNUM_X121 = 10  # X.121 (X.25, Frame Relay)
AFNUM_IPX = 11  # IPX
AFNUM_ATALK = 12  # Appletalk
AFNUM_DECNET = 13  # Decnet IV
AFNUM_BANYAN = 14  # Banyan Vines
AFNUM_E164NSAP = 15  # E.164 with NSAP format subaddress
AFNUM_DNS = 16  # DNS (Domain Name System)
AFNUM_DISTNAME = 17  # Distinguished Name
AFNUM_AS_NUMBER = 18  # AS Number
AFNUM_XTP_IP4 = 19  # XTP over IP version 4
AFNUM_XTP_IP6 = 20  # XTP over IP version 6
AFNUM_XTP = 21  # XTP native mode XTP
AFNUM_FC_WWPN = 22  # Fibre Channel World-Wide Port Name
AFNUM_FC_WWNN = 23  # Fibre Channel World-Wide Node Name
AFNUM_GWID = 24  # GWID
# draft-kompella-ppvpn-l2vpn
AFNUM_L2VPN = 25
AFNUM_L2VPN_OLD = 196
AFNUM_EIGRP_COMMON = 16384  # EIGRP Common Service Family
AFNUM_EIGRP_IPV4 = 16385  # EIGRP IPv4 Service Family
AFNUM_EIGRP_IPV6 = 16386  # EIGRP IPv6 Service Family
AFNUM_LCAF = 16387  # LISP Canonical Address Format
AFNUM_BGPLS = 16388  # draft BGP-LS



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" All BGP constant values """

# some handy things to know
BGP_MAX_PACKET_SIZE = 4096
BGP_MARKER_SIZE = 16  # size of BGP marker
BGP_HEADER_SIZE = 19  # size of BGP header, including marker
BGP_MIN_OPEN_MSG_SIZE = 29
BGP_MIN_UPDATE_MSG_SIZE = 23
BGP_MIN_NOTIFICATION_MSG_SIZE = 21
BGP_MIN_KEEPALVE_MSG_SIZE = BGP_HEADER_SIZE
BGP_TCP_PORT = 179
BGP_ROUTE_DISTINGUISHER_SIZE = 8

# BGP message types
BGP_OPEN = 1
BGP_UPDATE = 2
BGP_NOTIFICATION = 3
BGP_KEEPALIVE = 4
BGP_ROUTE_REFRESH = 5
BGP_CAPABILITY = 6
BGP_ROUTE_REFRESH_CISCO = 0x80

BGP_SIZE_OF_PATH_ATTRIBUTE = 2

# attribute flags, from RFC1771
BGP_ATTR_FLAG_OPTIONAL = 0x80
BGP_ATTR_FLAG_TRANSITIVE = 0x40
BGP_ATTR_FLAG_PARTIAL = 0x20
BGP_ATTR_FLAG_EXTENDED_LENGTH = 0x10


# SSA flags
BGP_SSA_TRANSITIVE = 0x8000
BGP_SSA_TYPE = 0x7FFF

# SSA Types
BGP_SSA_L2TPv3 = 1
BGP_SSA_mGRE = 2
BGP_SSA_IPSec = 3
BGP_SSA_MPLS = 4
BGP_SSA_L2TPv3_IN_IPSec = 5
BGP_SSA_mGRE_IN_IPSec = 6

# AS_PATH segment types
AS_SET = 1  # RFC1771
AS_SEQUENCE = 2  # RFC1771
AS_CONFED_SET = 4  # RFC1965 has the wrong values, corrected in
AS_CONFED_SEQUENCE = 3  # draft-ietf-idr-bgp-confed-rfc1965bis-01.txt

# OPEN message Optional Parameter types
BGP_OPTION_AUTHENTICATION = 1  # RFC1771
BGP_OPTION_CAPABILITY = 2  # RFC2842

# attribute types
BGPTYPE_ORIGIN = 1  # RFC1771
BGPTYPE_AS_PATH = 2  # RFC1771
BGPTYPE_NEXT_HOP = 3  # RFC1771
BGPTYPE_MULTI_EXIT_DISC = 4  # RFC1771
BGPTYPE_LOCAL_PREF = 5  # RFC1771
BGPTYPE_ATOMIC_AGGREGATE = 6  # RFC1771
BGPTYPE_AGGREGATOR = 7  # RFC1771
BGPTYPE_COMMUNITIES = 8  # RFC1997
BGPTYPE_ORIGINATOR_ID = 9  # RFC2796
BGPTYPE_CLUSTER_LIST = 10  # RFC2796
BGPTYPE_DPA = 11  # work in progress
BGPTYPE_ADVERTISER = 12  # RFC1863
BGPTYPE_RCID_PATH = 13  # RFC1863
BGPTYPE_MP_REACH_NLRI = 14  # RFC2858
BGPTYPE_MP_UNREACH_NLRI = 15  # RFC2858
BGPTYPE_EXTENDED_COMMUNITY = 16  # Draft Ramachandra
BGPTYPE_NEW_AS_PATH = 17  # draft-ietf-idr-as4bytes
BGPTYPE_NEW_AGGREGATOR = 18  # draft-ietf-idr-as4bytes
BGPTYPE_SAFI_SPECIFIC_ATTR = 19  # draft-kapoor-nalawade-idr-bgp-ssa-00.txt
BGPTYPE_TUNNEL_ENCAPS_ATTR = 23  # RFC5512
BGPTYPE_LINK_STATE = 99
BGPTYPE_ATTRIBUTE_SET = 128

#  VPN Route Target  #
BGP_EXT_COM_RT_0 = 0x0002  # Route Target,Format AS(2bytes):AN(4bytes)
BGP_EXT_COM_RT_1 = 0x0102  # Route Target,Format IPv4 address(4bytes):AN(2bytes)
BGP_EXT_COM_RT_2 = 0x0202  # Route Target,Format AS(4bytes):AN(2bytes)

# Route Origin (SOO site of Origin)
BGP_EXT_COM_RO_0 = 0x0003  # Route Origin,Format AS(2bytes):AN(4bytes)
BGP_EXT_COM_RO_1 = 0x0103  # Route Origin,Format IP address:AN(2bytes)
BGP_EXT_COM_RO_2 = 0x0203  # Route Origin,Format AS(2bytes):AN(4bytes)

# BGP Flow Spec
BGP_EXT_REDIRECT_NH = 0x0800  # redirect to ipv4/v6 nexthop
BGP_EXT_TRA_RATE = 0x8006  # traffic-rate 2-byte as#, 4-byte float
BGP_EXT_TRA_ACTION = 0x8007  # traffic-action bitmask
BGP_EXT_REDIRECT_VRF = 0x8008  # redirect 6-byte Route Target
BGP_EXT_TRA_MARK = 0x8009  # traffic-marking DSCP value

# BGP cost cummunity
BGP_EXT_COM_COST = 0x4301

# BGP link bandwith
BGP_EXT_COM_LINK_BW = 0x4004

# Unkonw
BGP_EXT_COM_UNKNOW = 0x0000

# route distinguisher type
BGP_ROUTE_DISTINGUISHER_TYPE_0 = 0x0000
BGP_ROUTE_DISTINGUISHER_TYPE_1 = 0x0001
BGP_ROUTE_DISTINGUISHER_TYPE_2 = 0x0001

# NLRI type as define in BGP flow spec RFC
BGPNLRI_FSPEC_DST_PFIX = 1  # RFC 5575
BGPNLRI_FSPEC_SRC_PFIX = 2  # RFC 5575
BGPNLRI_FSPEC_IP_PROTO = 3  # RFC 5575
BGPNLRI_FSPEC_PORT = 4  # RFC 5575
BGPNLRI_FSPEC_DST_PORT = 5  # RFC 5575
BGPNLRI_FSPEC_SRC_PORT = 6  # RFC 5575
BGPNLRI_FSPEC_ICMP_TP = 7  # RFC 5575
BGPNLRI_FSPEC_ICMP_CD = 8  # RFC 5575
BGPNLRI_FSPEC_TCP_FLAGS = 9  # RFC 5575
BGPNLRI_FSPEC_PCK_LEN = 10  # RFC 5575
BGPNLRI_FSPEC_DSCP = 11  # RFC 5575
BGPNLRI_FSPEC_FRAGMENT = 12  # RFC 5575

# BGP message Constants
VERSION = 4
PORT = 179
HDR_LEN = 19
MAX_LEN = 4096

# BGP messages type
MSG_BGP_CLOSED = 0
MSG_OPEN = 1
MSG_UPDATE = 2
MSG_NOTIFICATION = 3
MSG_KEEPALIVE = 4
MSG_ROUTEREFRESH = 5
MSG_CISCOROUTEREFRESH = 128

# BGP Capabilities Support

SUPPORT_4AS = False
CISCO_ROUTE_REFRESH = False
NEW_ROUTE_REFRESH = False
GRACEFUL_RESTART = False

# AFI_SAFI mapping

AFI_SAFI_DICT = {
    (1, 1): 'ipv4',
    (2, 1): 'ipv6',
    (1, 133): 'flowspec'
}
AFI_SAFI_STR_DICT = {
    'ipv6': (2, 1),
    'ipv4': (1, 1),
    'flowspec': (1, 133)
}

ADD_PATH_ACT_DICT = {
    1: 'receive',
    2: 'send',
    3: 'both'
}
# BGP FSM State
ST_IDLE = 1
ST_CONNECT = 2
ST_ACTIVE = 3
ST_OPENSENT = 4
ST_OPENCONFIRM = 5
ST_ESTABLISHED = 6

# BGP Timer (seconds)
DELAY_OPEN_TIME = 10
ROUTE_REFRESH_TIME = 10
LARGER_HOLD_TIME = 4 * 60
CONNECT_RETRY_TIME = 30
IDLEHOLD_TIME = 30
HOLD_TIME = 120

stateDescr = {
    ST_IDLE: "IDLE",
    ST_CONNECT: "CONNECT",
    ST_ACTIVE: "ACTIVE",
    ST_OPENSENT: "OPENSENT",
    ST_OPENCONFIRM: "OPENCONFIRM",
    ST_ESTABLISHED: "ESTABLISHED"
}

# Notification error codes
ERR_MSG_HDR = 1
ERR_MSG_OPEN = 2
ERR_MSG_UPDATE = 3
ERR_HOLD_TIMER_EXPIRED = 4
ERR_FSM = 5
ERR_CEASE = 6
ERR_CAP = 7

# Notification error codes dict
NOTIFICATION_ERROR_CODES_DICT = {
    ERR_MSG_HDR: "Message Header Error",
    ERR_MSG_OPEN: "OPEN Message Error",
    ERR_MSG_UPDATE: "UPDATE Message Error",
    ERR_HOLD_TIMER_EXPIRED: "Hold Timer Expired",
    ERR_FSM: "Finite State Machine Error",
    ERR_CEASE: "Cease",
    ERR_CAP: "CAPABILITY Message Error"
}

# Notification suberror codes
ERR_MSG_HDR_CONN_NOT_SYNC = 1
ERR_MSG_HDR_BAD_MSG_LEN = 2
ERR_MSG_HDR_BAD_MSG_TYPE = 3

ERR_MSG_OPEN_UNSUP_VERSION = 1
ERR_MSG_OPEN_BAD_PEER_AS = 2
ERR_MSG_OPEN_BAD_BGP_ID = 3
ERR_MSG_OPEN_UNSUP_OPT_PARAM = 4
ERR_MSG_OPEN_UNACCPT_HOLD_TIME = 6
ERR_MSG_OPEN_UNSUP_CAPA = 7  # RFC 5492
ERR_MSG_OPEN_UNKNO = 8

ERR_MSG_UPDATE_MALFORMED_ATTR_LIST = 1
ERR_MSG_UPDATE_UNRECOGNIZED_WELLKNOWN_ATTR = 2
ERR_MSG_UPDATE_MISSING_WELLKNOWN_ATTR = 3
ERR_MSG_UPDATE_ATTR_FLAGS = 4
ERR_MSG_UPDATE_ATTR_LEN = 5
ERR_MSG_UPDATE_INVALID_ORIGIN = 6
ERR_MSG_UPDATE_INVALID_NEXTHOP = 8
ERR_MSG_UPDATE_OPTIONAL_ATTR = 9
ERR_MSG_UPDATE_INVALID_NETWORK_FIELD = 10
ERR_MSG_UPDATE_MALFORMED_ASPATH = 11
ERR_MSG_UPDATE_UNKOWN_ATTR = 12

NOTIFICATION_SUB_ERROR_CODES_DICT = {
    ERR_MSG_HDR: {
        ERR_MSG_HDR_CONN_NOT_SYNC: 'Connection Not Synchronized',  # 1
        ERR_MSG_HDR_BAD_MSG_LEN: 'Bad Message Length',             # 2
        ERR_MSG_HDR_BAD_MSG_TYPE: 'Bad Message Type'               # 3
    },
    ERR_MSG_OPEN: {
        ERR_MSG_OPEN_UNSUP_VERSION: 'Unsupported Version Number',
        ERR_MSG_OPEN_BAD_PEER_AS: 'Bad Peer AS',
        ERR_MSG_OPEN_BAD_BGP_ID: 'Bad BGP Identifier',
        ERR_MSG_OPEN_UNSUP_OPT_PARAM: 'Unsupported Optional Parameter',
        ERR_MSG_OPEN_UNACCPT_HOLD_TIME: 'Unacceptable Hold Time',
        ERR_MSG_OPEN_UNSUP_CAPA: 'Unsupported Capability',
        ERR_MSG_OPEN_UNKNO: 'NULL',
    },
    ERR_MSG_UPDATE: {
        ERR_MSG_UPDATE_MALFORMED_ATTR_LIST: 'Malformed Attribute List',
        ERR_MSG_UPDATE_UNRECOGNIZED_WELLKNOWN_ATTR: 'Unrecognized Well-known Attribute',
        ERR_MSG_UPDATE_MISSING_WELLKNOWN_ATTR: 'Missing Well-known Attribute',
        ERR_MSG_UPDATE_ATTR_FLAGS: 'Attribute Flags Error',
        ERR_MSG_UPDATE_ATTR_LEN: 'Attribute Length Error',
        ERR_MSG_UPDATE_INVALID_ORIGIN: 'Invalid ORIGIN Attribute',
        ERR_MSG_UPDATE_INVALID_NEXTHOP: 'Invalid NEXT_HOP Attribute',
        ERR_MSG_UPDATE_OPTIONAL_ATTR: 'Optional Attribute Error',
        ERR_MSG_UPDATE_INVALID_NETWORK_FIELD: 'Invalid Network Field',
        ERR_MSG_UPDATE_MALFORMED_ASPATH: 'Malformed AS_PATH',
        ERR_MSG_UPDATE_UNKOWN_ATTR: 'NULL'
    }

}

ATTRIBUTE_ID_2_STR = {
    1: 'ORIGIN',
    2: 'AS_PATH',
    3: 'NEXT_HOP',
    4: 'MULTI_EXIT_DISC',
    5: 'LOCAL_PREF',
    6: 'ATOMIC_AGGREGATE',
    7: 'AGGREGATOR',
    8: 'COMMUNITY',
    9: 'ORIGINATOR_ID',
    10: 'CLUSTER_LIST',
    14: 'MP_REACH_NLRI',
    15: 'MP_UNREACH_NLRI',
    16: 'EXTENDED_COMMUNITY',
    17: 'AS4_PATH',
    18: 'AS4_AGGREGATOR'
}

ATTRIBUTE_STR_2_ID = dict([(k, v) for (k, v) in ATTRIBUTE_ID_2_STR.items()])


WELL_KNOW_COMMUNITY_INT_2_STR = {
    0xFFFF0000: 'PLANNED_SHUT',
    0xFFFF0001: 'ACCEPT_OWN',
    0xFFFF0002: 'ROUTE_FILTER_TRANSLATED_v4',
    0xFFFF0003: 'ROUTE_FILTER_v4',
    0xFFFF0004: 'ROUTE_FILTER_TRANSLATED_v6',
    0xFFFF0005: 'ROUTE_FILTER_v6',
    0xFFFFFF01: 'NO_EXPORT',
    0xFFFFFF02: 'NO_ADVERTISE',
    0xFFFFFF03: 'NO_EXPORT_SUBCONFED',
    0xFFFFFF04: 'NOPEER'
}

WELL_KNOW_COMMUNITY_STR_2_INT = dict(
    [(r, l) for (l, r) in WELL_KNOW_COMMUNITY_INT_2_STR.items()])

TCP_MD5SIG_MAXKEYLEN = 80
SS_PADSIZE_IPV4 = 120
TCP_MD5SIG = 14
SS_PADSIZE_IPV6 = 100
SIN6_FLOWINFO = 0
SIN6_SCOPE_ID = 0


COMMUNITY_DICT = False



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test app"""

import unittest
from yabgp.api.app import app


class FlaskrTestCase(unittest.TestCase):

    def setUp(self):
        # prepare some peers
        self.app = app.test_client()

    def tearDown(self):
        """"""
        pass

    def test_show_bgp_summary(self):
        result = self.app.get('/')
        self.assertTrue(result)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" mongodb """

import logging

import pymongo

LOG = logging.getLogger(__name__)


class MongoApi(object):
    """Class handling MongoDB specific functionality.

    This class uses PyMongo APIs internally to create database connection,
    also serves as handle to collection for get/set operations, etc.

    """
    # class level attributes for re-use of db client connection and collection
    _DB = {}  # dict of db_name: db connection reference
    _MONGO_COLLS = {}  # dict of cache_collection : db collection reference

    def __init__(self, connection_url, db_name, use_replica=False, replica_name='rs1',
                 read_preference=3, write_concern=-1, w_timeout=5000):
        """for parameters details, please see
        http://api.mongodb.org/python/current/api/pymongo/mongo_client.html

        :param connection_url: string like mongodb://localhost:27017
        :param db_name: database name, string
        :param use_replica: if use replica, true of false
        :param replica_name: if use replica, then the replica set name
        :param read_preference: read preference, interger
        :param write_concern: write concern, interger
        :param w_timeout: write concern timeout, interger
        """
        LOG.debug('Creating MongoDB API class')
        self.conection_url = connection_url
        self.db_name = db_name
        self.collection_name = None

        # config about replica set
        self.use_replica = use_replica
        self.replica_name = replica_name
        self.read_preference = read_preference

        # Write Concern options:
        self.w = write_concern
        self.w_timeout = w_timeout

    def _get_db(self):
        """
        get database
        :return: database
        """
        try:
            if self.use_replica:  # if use replica set configuration
                connection = pymongo.MongoClient(
                    self.conection_url, replicaSet=self.replica_name)
            else:  # use for standalone node or mongos in sharded setup
                connection = pymongo.MongoClient(self.conection_url)

        except pymongo.errors.ConnectionFailure as e:
            LOG.warn('Unable to connect to the database server, %s', e)
            raise

        database = getattr(connection, self.db_name)
        return database

    def _close_db(self):
        if self.db_name in self._DB:
            self._DB[self.db_name].client.close()
            self._DB.pop(self.db_name)

    def get_collection(self):
        """
        get collection
        :return: database collection
        """
        if self.collection_name not in self._MONGO_COLLS:

            if self.db_name not in self._DB:
                self._DB[self.db_name] = self._get_db()
            coll = getattr(self._DB[self.db_name], self.collection_name)

            if self.use_replica:
                # for read preference
                # TODO(xiaoquwl) fix this as more elegant way in future
                if self.read_preference is not None:
                    if self.read_preference == 0:
                        coll.with_options(read_preference=pymongo.ReadPreference.PRIMARY)
                    elif self.read_preference == 1:
                        coll.with_options(read_preference=pymongo.ReadPreference.PRIMARY_PREFERRED)
                    elif self.read_preference == 2:
                        coll.with_options(read_preference=pymongo.ReadPreference.SECONDARY)
                    elif self.read_preference == 3:
                        coll.with_options(read_preference=pymongo.ReadPreference.SECONDARY_PREFERRED)
                    elif self.read_preference == 4:
                        coll.with_options(read_preference=pymongo.ReadPreference.NEAREST)
                    else:
                        LOG.error('unknow read preference setting')
                        pass

                # for write concern
                if self.w > -1:
                    coll.write_concern['w'] = self.w
                    coll.write_concern['wtimeout'] = 5000

            self._MONGO_COLLS[self.collection_name] = coll

        return self._MONGO_COLLS[self.collection_name]

    def remove_collection(self):
        self._DB[self.db_name].drop_collection(self.collection_name)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Community attribute """

import unittest

from yabgp.common import exception as excep
from yabgp.common.constants import ERR_MSG_UPDATE_ATTR_LEN
from yabgp.message.attribute.community import Community


class TestCommunity(unittest.TestCase):

    def test_parse(self):

        community = Community.parse(value=b'\xff\xff\xff\x01')
        self.assertEqual(['NO_EXPORT'], community)

        community = Community.parse(value=b'\xff\xff\xff\x01\x12\xe5&\xc9')
        self.assertEqual(['NO_EXPORT', '4837:9929'], community)

        community = Community.parse(value=b'\x12\xe5\x04\xd7\x12\xe5&\xc9')
        self.assertEqual(['4837:1239', '4837:9929'], community)

        self.assertRaises(excep.UpdateMessageError, Community.parse,
                          value=b'\xff\xff\xff\x01\x01')
        try:
            Community.parse(value=b'\xff\xff\xff\x01\x01')
        except excep.UpdateMessageError as e:
            self.assertEqual(ERR_MSG_UPDATE_ATTR_LEN, e.sub_error)

    def test_construct(self):

        community = Community.construct(value=['NO_EXPORT'])
        self.assertEqual(b'\xc0\x08\x04\xff\xff\xff\x01', community)

        community = Community.construct(value=['NO_EXPORT', '4837:9929'])
        self.assertEqual(b'\xc0\x08\x08\xff\xff\xff\x01\x12\xe5&\xc9', community)

        community = Community.construct(value=['4837:1239', '4837:9929'])
        self.assertEqual(b'\xc0\x08\x08\x12\xe5\x04\xd7\x12\xe5&\xc9', community)

    def test_construct_invalid(self):
        self.assertRaises(excep.UpdateMessageError, Community.construct, value=['111:111', 'a:b'])

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2007 by Mark Bergsma <mark@nedworks.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP FMS Implementation"""

import time
import logging

from yabgp.core.timer import BGPTimer
from yabgp.common import constants as bgp_cons

LOG = logging.getLogger(__name__)


class FSM(object):

    """
    Implements BGP Events described in section 8.1 of RFC 4271
    Implements BGP finite state machine described in section 8.2 of RFC 4271
    """

    protocol = None
    state = bgp_cons.ST_IDLE
    large_hold_time = bgp_cons.LARGER_HOLD_TIME

    def __init__(self, bgp_peering=None, protocol=None):

        """
        please see RFC 4271 page 37 for value meanning
        """

        self.bgp_peering = bgp_peering
        self.protocol = protocol

        # Session attributes required (mandatory) for each connection:
        self.state = bgp_cons.ST_IDLE
        self.connect_retry_counter = 0
        self.connect_retry_time = bgp_cons.CONNECT_RETRY_TIME
        self.connect_retry_timer = BGPTimer(self.connect_retry_time_event)
        self.hold_time = bgp_cons.HOLD_TIME

        self.hold_timer = BGPTimer(self.hold_time_event)
        self.keep_alive_time = self.hold_time / 3
        self.keep_alive_timer = BGPTimer(self.keep_alive_time_event)

        self.allow_automatic_start = True
        self.allow_automatic_stop = False
        self.delay_open = False
        self.delay_open_time = bgp_cons.DELAY_OPEN_TIME
        self.delay_open_timer = BGPTimer(self.delay_open_time_event)
        self.idle_hold_time = bgp_cons.IDLEHOLD_TIME
        self.idle_hold_timer = BGPTimer(self.idle_hold_time_event)

        self.uptime = None

    def __setattr__(self, name, value):

        if name == 'state' and value != getattr(self, name):
            LOG.info("[%s]State is now:%s", self.bgp_peering.peer_addr, bgp_cons.stateDescr[value])
            if value == bgp_cons.ST_ESTABLISHED:
                self.uptime = time.time()
        super(FSM, self).__setattr__(name, value)

    def manual_start(self):

        """
        Event 1: ManualStart
        Definiton: Should be called when a BGP ManualStart event is requested.
                   Note that a protocol instance does not yet exist at this point,
                   so this method requires some support from BGPPeering.manual_start().
        Status: Mandatory
        """
        LOG.info('Manual start.')
        if self.state == bgp_cons.ST_IDLE:
            self.connect_retry_counter = 0
            self.connect_retry_timer.reset(self.connect_retry_time)

    def manual_stop(self):

        """
        Event 2: ManualStop
        Definition: Should be called when a BGP ManualStop event is requested.
        Status: Mandatory
        """
        LOG.info('Manual stop')
        if self.state != bgp_cons.ST_IDLE:

            self.protocol.send_notification(bgp_cons.ERR_CEASE, 0)
            # Stop all timers
            LOG.info('Stop all timers')
            for timer in (self.connect_retry_timer, self.hold_timer, self.keep_alive_timer,
                          self.delay_open_timer, self.idle_hold_timer):
                timer.cancel()
                LOG.info('-- Stop timer %s', timer)
            self._close_connection()
            self.connect_retry_counter = 0
            self.allow_automatic_start = False
            self.state = bgp_cons.ST_IDLE

    def automatic_start(self, idle_hold=False):

        """
        Event 3: AutomaticStart
        Definition: Should be called when a BGP Automatic Start event is requested.
                    Returns True or False to indicate BGPPeering whether a connection
                    attempt should be initiated.
        Status: Optional

        :param idle_hold: BGP Idle Hold or not.
        """
        LOG.info('Automatic start')
        if self.state in [bgp_cons.ST_IDLE, bgp_cons.ST_CONNECT]:
            if idle_hold:
                LOG.info('Idle Hold, please wait time=%s', self.idle_hold_time)
                self.idle_hold_timer.reset(self.idle_hold_time)
                return False
            elif self.allow_automatic_start:
                LOG.info('Do not need Idle Hold, start right now.')
                LOG.info('Connect retry counter: %s', self.connect_retry_counter)
                self.connect_retry_counter += 1
                self.connect_retry_timer.reset(self.connect_retry_time)
                LOG.info('Connect retry timer, time=%s', self.connect_retry_time)
                self.state = bgp_cons.ST_CONNECT
                return True
            else:
                return False
        else:
            return False

    def connect_retry_time_event(self):

        """
        Event 9: ConnectRetryTimer_Expires
        Definition: Called when the ConnectRetryTimer expires.
        Status: Mandatory
        """
        LOG.info('Connect retry timer expires')
        if self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            # State Connect, event 9
            self._close_connection()
            LOG.info('Reset connect retry timer, time=%s', self.connect_retry_time)
            self.connect_retry_timer.reset(self.connect_retry_time)
            self.delay_open_timer.cancel()
            LOG.info('Cancel delay open timer')
            # Initiate TCP connection
            if self.bgp_peering:
                LOG.info('Bgp peering connect retry')
                self.bgp_peering.connect_retry()
        elif self.state != bgp_cons.ST_IDLE:
            # State OpenSent, OpenConfirm, Established, event 12
            self.protocol.send_notification(bgp_cons.ERR_FSM, 0)
            self._error_close()

    def hold_time_event(self):

        """
        Event 10:   HoldTimer_Expires
        Definition: Called when the HoldTimer expires.
        Action:     NOTIFICATION message with the Hold Timer Expried Error
                    Code is sent and the BGP connection is closed
        Status:     Mandatory
        """
        LOG.info('Hold Timer expires')
        if self.state in (bgp_cons.ST_OPENSENT, bgp_cons.ST_OPENCONFIRM, bgp_cons.ST_ESTABLISHED):
            # States OpenSent, OpenConfirm, Established, event 10
            self.protocol.send_notification(bgp_cons.ERR_HOLD_TIMER_EXPIRED, 0)
            self.connect_retry_timer.cancel()
            self._error_close()
            self.state = bgp_cons.ST_IDLE
        elif self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            self._error_close()

    def keep_alive_time_event(self):

        """
        Event 11:   KeepaliveTimer_Expires
        Definition: Called when the KeepAliveTimer expires.
        Action:     Send Keepalive messsage if the state of FSM is ST_OPENCONFIRM
                    or ST_ESTABLISHED, close BGP connection if state is others
        Status: Mandatory
        """
        LOG.info('Keep alive timer expires')
        if self.state in (bgp_cons.ST_OPENCONFIRM, bgp_cons.ST_ESTABLISHED):
            # State OpenConfirm, Established, event 11
            self.protocol.send_keepalive()
            if self.hold_time > 0:
                self.keep_alive_timer.reset(self.keep_alive_time)
        elif self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            self._error_close()

    def delay_open_time_event(self):

        """
        Event 12: DelayOpenTimer_Expires
        Definition: Called when the DelayOpenTimer expires.
        Status: Optional
        """
        LOG.info('Delay open timer expires')
        # DelayOpen attribute SHOULD be set to TRUE

        if self.state == bgp_cons.ST_CONNECT:
            # State Connect, event 12
            self.protocol.send_open()
            self.hold_timer.reset(self.large_hold_time)
            self.state = bgp_cons.ST_OPENSENT
        elif self.state == bgp_cons.ST_ACTIVE:
            # State Active, event 12
            self.connect_retry_timer.cancel()
            self.delay_open_timer.cancel()
            self.protocol.send_open()
            self.hold_timer.reset(self.large_hold_time)
            self.state = bgp_cons.ST_OPENSENT
        elif self.state != bgp_cons.ST_IDLE:
            # State OpenSent, OpenConfirm, Established, event 12
            self.protocol.send_notification(bgp_cons.ERR_FSM, 0)
            self._error_close()

    def idle_hold_time_event(self):

        """
        Event 13: IdleHoldTimer_Expires
        Definition: Called when the IdleHoldTimer expires.
        Status: Optional
        """
        # DampPeerOscillations attribute SHOULD be set to TRUE
        # assert(self.dampPeerOscillations)
        LOG.info('Idle Hold Timer expires')
        if self.state == bgp_cons.ST_IDLE:
            if self.bgp_peering:
                self.bgp_peering.automatic_start(idle_hold=False)

    def connection_made(self):

        """
        Event 16: Tcp_CR_Acked
        Event 17: TcpConnectionConfirmed
        Definition: Should be called when a TCP connection has successfully been
                    established with the peer.
        Status: Mandatory
        """
        if self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            # State Connect, Event 16 or 17
            if self.delay_open:
                self.connect_retry_timer.cancel()
                LOG.info('Delay open for this peer')
                self.delay_open_timer.reset(self.delay_open_time)
            else:
                self.connect_retry_timer.cancel()
                self.protocol.send_open()
                self.hold_timer.reset(self.large_hold_time)
                self.state = bgp_cons.ST_OPENSENT

    def connection_failed(self):

        """
        Event 18: TcpConnectionFails
        Definition: Should be called when the associated TCP connection failed,
                    or was lost.
        Status: Mandatory
        """
        if self.state == bgp_cons.ST_CONNECT:
            # State Connect, event 18
            if self.delay_open_timer.active():
                self.connect_retry_timer.reset(self.connect_retry_time)
                self.delay_open_timer.cancel()
                self.state = bgp_cons.ST_ACTIVE
            else:
                self.connect_retry_timer.cancel()
                self._close_connection()
                if self.bgp_peering:
                    self.state = bgp_cons.ST_IDLE
                    self.bgp_peering.connection_closed(self.protocol)
        elif self.state == bgp_cons.ST_ACTIVE:
            # State Active, event 18
            self.connect_retry_timer.reset(self.connect_retry_time)
            self.delay_open_timer.cancel()
            if self.bgp_peering:
                # self.bgp_peering.releaseResources(self.protocol)
                pass
            # TODO: osc damping
            self.state = bgp_cons.ST_IDLE
        elif self.state == bgp_cons.ST_OPENSENT:
            # State OpenSent, event 18
            if self.bgp_peering:
                # self.bgp_peering.releaseResources(self.protocol)
                pass
            self._close_connection()
            self.connect_retry_timer.reset(self.connect_retry_time)
            self.state = bgp_cons.ST_ACTIVE
            if self.bgp_peering:
                self.bgp_peering.connection_closed(self.protocol)
        elif self.state in (bgp_cons.ST_OPENCONFIRM, bgp_cons.ST_ESTABLISHED):
            self._error_close()

    def open_received(self):

        """
        Event 19: BGPOpen
        Event 20: BGPOpen with DelayOpenTimer running
        Definition: Should be called when a BGP Open message was
                    received from the peer.
        Status: Mandatory
        """

        if self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            if self.delay_open_timer.active():
                # State Connect, event 20
                self.connect_retry_timer.cancel()
                self.delay_open_timer.cancel()
                self.connect_retry_counter = 0
                self.protocol.send_open()
                self.protocol.send_keepalive()
                if self.hold_time:
                    self.keep_alive_timer.reset(self.keep_alive_time)
                    self.hold_timer.reset(self.hold_time)
                else:    # holdTime == 0
                    self.keep_alive_timer.cancel()
                    self.hold_timer.cancel()

                self.state = bgp_cons.ST_OPENCONFIRM
            else:
                # State Connect, event 19
                self._error_close()

        elif self.state == bgp_cons.ST_OPENSENT:
            # State OpenSent, events 19, 20
            self.delay_open_timer.cancel()
            self.connect_retry_timer.cancel()
            self.protocol.send_keepalive()
            if self.hold_time > 0:
                self.keep_alive_timer.reset(self.keep_alive_time)
                self.hold_timer.reset(self.hold_time)
            self.state = bgp_cons.ST_OPENCONFIRM

        elif self.state == bgp_cons.ST_OPENCONFIRM:
            # State OpenConfirm, events 19, 20
            # TODO:Perform collision detection
            pass

        elif self.state == bgp_cons.ST_ESTABLISHED:
            # State Established, event 19 or 20
            self.protocol.send_notification(bgp_cons.ERR_FSM, 0)
            self._error_close()

    def header_error(self, suberror, data=b''):

        """
        Event 21: BGPHeaderErr
        Definition: Should be called when an invalid BGP
                    message header was received.
        Status: Mandatory

        :param suberror: bgp notification sub error code
        :param data: bgp notification error data
        """

        self.protocol.send_notification(bgp_cons.ERR_MSG_HDR, suberror, data)
        # Note: RFC4271 states that we should send ERR_FSM in the
        # Established state, which contradicts earlier statements.
        self._error_close()

    def open_message_error(self, suberror, data=b''):

        """
        Event 22: BGPOpenMsgErr
        Definition: Should be called when an invalid BGP
                    Open message was received.
        Status: Mandatory

        :param suberror: bgp notification sub error code
        :param data: bgp notification error data
        """

        self.protocol.send_notification(bgp_cons.ERR_MSG_OPEN, suberror, data)
        # Note: RFC4271 states that we should send ERR_FSM in the
        # Established state, which contradicts earlier statements.
        self._error_close()

    def notimsg_version_error(self):

        """
        Event 24: NotifMsgVerErr
        Definition: Should be called when a BGP Notification Open Version
                    Error message was received from the peer.
        Status: Mandatory
        """

        if self.state in (bgp_cons.ST_OPENSENT, bgp_cons.ST_OPENCONFIRM):
            # State OpenSent, event 24
            self.connect_retry_timer.cancel()
            self._close_connection()
            self.state = bgp_cons.ST_IDLE
        elif self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            # State Connect, event 24
            self._error_close()

    def notification_received(self, error, suberror):

        """
        Event 25: NotifMsg
        Definition: Should be called when a BGP Notification message was
                    received from the peer.
        Status: Mandatory

        :param error: bgp notification error code
        :param suberror: bgp notification sub error code
        """

        if error == bgp_cons.ERR_MSG_OPEN and suberror == 1:
            # Event 24 : version error
            self.notimsg_version_error()
        else:
            if self.state != bgp_cons.ST_IDLE:
                # State != Idle, events 24, 25
                self._error_close()

    def keep_alive_received(self):

        """
        Event 26: KeepAliveMsg
        Definition: Should be called when a BGP KeepAlive packet
                    was received from the peer.
        Status: Mandatory
        """

        if self.state == bgp_cons.ST_OPENCONFIRM:
            # State OpenSent, event 26
            self.hold_timer.reset(self.hold_time)
            self.state = bgp_cons.ST_ESTABLISHED
        elif self.state == bgp_cons.ST_ESTABLISHED:
            # State Established, event 26
            self.hold_timer.reset(self.hold_time)
        elif self.state in (bgp_cons.ST_CONNECT, bgp_cons.ST_ACTIVE):
            # States Connect, Active, event 26
            self._error_close()

    def update_received(self):

        """
        Event 27: UpdateMsg
        Definition: Called when a valid BGP Update message was received.
        Status: Mandatory
        """

        if self.state == bgp_cons.ST_ESTABLISHED:
            # State Established, event 27
            if self.hold_time != 0:
                self.hold_timer.reset(self.hold_time)

        elif self.state in (bgp_cons.ST_ACTIVE, bgp_cons.ST_CONNECT):
            # States Active, Connect, event 27
            self._error_close()
        elif self.state in (bgp_cons.ST_OPENSENT, bgp_cons.ST_OPENCONFIRM):
            # States OpenSent, OpenConfirm, event 27
            self.protocol.send_notification(bgp_cons.ERR_FSM, 0)
            self._error_close()

    def update_sent(self):

        """Called by the protocol instance when it just sent an Update message."""

        if self.hold_time > 0:
            self.keep_alive_timer.reset(self.keep_alive_time)

    def _error_close(self):

        """Internal method that closes a connection and returns the state
        to IDLE.
        """

        # Stop the timers
        for timer in (self.connect_retry_timer, self.delay_open_timer, self.hold_timer, self.keep_alive_timer):
            timer.cancel()

        self.idle_hold_timer.reset(self.idle_hold_time)

        # Release BGP resources (routes, etc)
        if self.bgp_peering:
            # self.bgp_peering.releaseResources(self.protocol)
            pass

        self._close_connection()

        self.connect_retry_counter += 1
        self.state = self.bgp_peering.fsm.state = bgp_cons.ST_IDLE

    def _close_connection(self):

        """Internal method that close the connection if a valid BGP protocol
        instance exists.
        """
        if self.protocol is not None:
            self.protocol.closeConnection()
            self.connect_retry_counter = 0
            LOG.info('Closing protocol connection.')



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Cluster List attribute """

import unittest

from yabgp.common.constants import ERR_MSG_UPDATE_ATTR_LEN
from yabgp.common import exception as excep
from yabgp.message.attribute.clusterlist import ClusterList


class TestClusterList(unittest.TestCase):

    def test_parse(self):

        cluster_list = ClusterList.parse(value=b'\x01\x01\x01\x01\x02\x02\x02\x02\x03\x03\x03\x03')
        self.assertEqual(['1.1.1.1', '2.2.2.2', '3.3.3.3'], cluster_list)

    def test_parse_invalid_length(self):
        # invalid length
        self.assertRaises(excep.UpdateMessageError, ClusterList.parse,
                          b'\x01\x01\x01\x01\x01\x02\x02\x02\x02\x03\x03\x03\x03')
        try:
            ClusterList.parse(value=b'\x01\x01\x01\x01\x01\x02\x02\x02\x02\x03\x03\x03\x03')
        except excep.UpdateMessageError as e:
            self.assertEqual(ERR_MSG_UPDATE_ATTR_LEN, e.sub_error)

    def test_construct(self):

        cluster_list = ClusterList.construct(value=['1.1.1.1', '2.2.2.2', '3.3.3.3'])
        self.assertEqual(b'\x80\n\x0c\x01\x01\x01\x01\x02\x02\x02\x02\x03\x03\x03\x03', cluster_list)

    def test_construct_invalid_length(self):
        # invalid length
        self.assertRaises(excep.UpdateMessageError, ClusterList.construct,
                          ['a', 2, '3'])

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" message queue config """

from oslo_config import cfg

from yabgp.channel.filter import FILTER_TYPR_INIT_DICT

CONF = cfg.CONF

rabbit_mq = [
    cfg.StrOpt('rabbit_host',
               default='localhost',
               help='The RabbitMQ broker address where a single node is used'),
    cfg.IntOpt('rabbit_port',
               default=5672,
               help='The RabbitMQ broker port where a single node is used'),
    cfg.BoolOpt('rabbit_use_ssl',
                default=False,
                help='Connect over SSL for RabbitMQ.'),
    cfg.StrOpt('rabbit_userid',
               default='guest',
               help='The RabbitMQ userid.'),
    cfg.StrOpt('rabbit_password',
               default='guest',
               help='The RabbitMQ password.',
               secret=True),
]

channle_filter = [
    cfg.DictOpt('filter',
                default=FILTER_TYPR_INIT_DICT,
                help='the community and prefix filter dict'),
]



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import os
import logging
import contextlib
import traceback

from oslo_config import cfg
from twisted.internet import reactor
from twisted.web.server import Site

from yabgp import version, log
from yabgp.core.factory import BGPPeering
from yabgp.config import get_bgp_config
from yabgp.common import constants as bgp_cons
from yabgp.api.app import app
from yabgp.channel.config import rabbit_mq
from yabgp.channel.config import channle_filter
from yabgp.channel.factory import PikaFactory
from yabgp.db import config as db_config
from yabgp.db.mongodb import MongoApi
from yabgp.db import constants as db_cons


log.early_init_log(logging.DEBUG)

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


@contextlib.contextmanager
def mongo_operation(mongo_conn, connection_name):
    mongo_conn.collection_name = connection_name
    db = mongo_conn.get_collection()
    yield db
    mongo_conn._close_db()


def load_channel_filter_from_db(peer_ip, mongo_api):
    """
    load rabbitmq channle filter from mongodb
    :return:
    """
    LOG.info('try to load yabgp rabbitmq channel filter from mongodb')
    mongo_api.collection_name = db_cons.MONGO_COLLECTION_RABBIT_CHANNEL_FILTER
    try:
        filter_list = mongo_api.get_collection().find({'peer_ip': peer_ip})
        for item in filter_list:
            if item['value'] not in CONF.rabbit_mq.filter[item['type']]:
                CONF.rabbit_mq.filter[item['type']][item['value']] = None
    except Exception as e:
        LOG.debug(traceback.format_exc())
        LOG.error('load failed, %s', e)
        sys.exit()
    pass


def load_bgp_policy_from_db(mongo_conn, connection_name):
    """
    load bgp policy from mongodb
    :return:
    """
    pass


def check_running_mode():
    """
    before start the bgp peering, we should check the running mode
    :return:
    """

    if not CONF.standalone:
        # not standalone?
        CONF.register_opts(rabbit_mq, group='rabbit_mq')
        CONF.register_opts(channle_filter, group='rabbit_mq')
        db_config.register_options()


def check_msg_config():

    LOG.info('Check configurations about message process')
    if CONF.message.write_disk:
        if not os.path.exists(CONF.message.write_dir):
            os.makedirs(CONF.message.write_dir)
            LOG.info('Create dir %s', CONF.message.write_dir)
        CONF.message.write_msg_max_size = CONF.message.write_msg_max_size * 1024 * 1024


def register_to_db(peer_ip, mongo_api):
    """
    register peer configuration to database
    :return:
    """
    LOG.info('try to register yabgp agent to database')
    peer_config = {
        '_id': '%s:%s:%s' % (CONF.rest.bind_host, CONF.rest.bind_port, peer_ip),
        'peer_ip': peer_ip,
        'bind_host': CONF.rest.bind_host,
        'bind_port': CONF.rest.bind_port,
        'local_as': CONF.bgp.running_config[peer_ip]['local_as'],
        'local_addr': CONF.bgp.running_config[peer_ip]['local_addr'],
        'remote_as': CONF.bgp.running_config[peer_ip]['remote_as'],
        'remote_addr': CONF.bgp.running_config[peer_ip]['remote_addr'],
        'afi_safi': CONF.bgp.afi_safi,
        'tag': CONF.bgp.running_config[peer_ip]['tag']

    }
    mongo_api.collection_name = db_cons.MONGO_COLLECTION_BGP_AGENT
    try:
        mongo_api.get_collection().save(peer_config)
    except Exception as e:
        LOG.debug(traceback.format_exc())
        LOG.error('register failed, %s', e)
        sys.exit()


def prepare_twisted_service():
    LOG.info('Prepare twisted services')
    # check all peers
    all_peers = {}

    # check running mode
    if not CONF.standalone:
        # rabbitmq factory
        rabbit_mq_factory = PikaFactory(
            host=CONF.rabbit_mq.rabbit_host,
            port=CONF.rabbit_mq.rabbit_port,
            userid=CONF.rabbit_mq.rabbit_userid,
            password=CONF.rabbit_mq.rabbit_password
        )
        rabbit_mq_factory.peer_list = CONF.bgp.running_config.keys()
        rabbit_mq_factory.connect()
        # mongodb connection
        if CONF.database.use_replica:
            mongo_connection = MongoApi(
                connection_url=CONF.database.connection,
                db_name=CONF.database.dbname,
                use_replica=CONF.database.use_replica,
                replica_name=CONF.database.replica_name,
                read_preference=CONF.database.read_preference,
                write_concern=CONF.database.write_concern,
                w_timeout=CONF.database.write_concern_timeout
            )
        else:
            mongo_connection = MongoApi(connection_url=CONF.database.connection, db_name=CONF.database.dbname)
        # check api bind host
        if CONF.rest.bind_host == '0.0.0.0':
            LOG.error('please use the exactly rest host ip address when not running in standalone mode')
            sys.exit()
        # TODO load channel filter and peer policy
    else:
        rabbit_mq_factory = None
        mongo_connection = None
    for peer in CONF.bgp.running_config:
        LOG.info('Get peer %s configuration', peer)
        if not CONF.standalone:
            if CONF.bgp.running_config[peer]['local_addr'] == '0.0.0.0':
                LOG.error('please use the exactly local bgp ip address when not running in standalone mode')
                sys.exit()
        if CONF.message.write_disk:
            msg_file_path_for_peer = os.path.join(
                CONF.message.write_dir,
                peer.lower()
            )
            if not os.path.exists(msg_file_path_for_peer):
                os.makedirs(msg_file_path_for_peer)
                LOG.info('Create dir %s for peer %s', msg_file_path_for_peer, peer)
            LOG.info('BGP message file path is %s', msg_file_path_for_peer)
        else:
            msg_file_path_for_peer = None
        LOG.info('Create BGPPeering instance')
        afi_safi_list = [bgp_cons.AFI_SAFI_STR_DICT[afi_safi]
                         for afi_safi in CONF.bgp.running_config[peer]['afi_safi']]
        CONF.bgp.running_config[peer]['afi_safi'] = afi_safi_list
        CONF.bgp.running_config[peer]['capability']['local']['afi_safi'] = afi_safi_list
        bgp_peering = BGPPeering(
            myasn=CONF.bgp.running_config[peer]['local_as'],
            myaddr=CONF.bgp.running_config[peer]['local_addr'],
            peerasn=CONF.bgp.running_config[peer]['remote_as'],
            peeraddr=CONF.bgp.running_config[peer]['remote_addr'],
            tag=CONF.bgp.running_config[peer]['tag'],
            afisafi=CONF.bgp.running_config[peer]['afi_safi'],
            msgpath=msg_file_path_for_peer,
            md5=CONF.bgp.running_config[peer]['md5'],
            channel=rabbit_mq_factory,
            mongo_conn=mongo_connection
        )
        all_peers[peer] = bgp_peering
        CONF.bgp.running_config[peer]['factory'] = bgp_peering

        # register to database and check agent role
        if not CONF.standalone:
            register_to_db(peer_ip=peer, mongo_api=mongo_connection)
            if not CONF.bgp.running_config[peer]['tag']:
                LOG.error('Please point out the role tag(SRC,DST or BOTH)for not running in standalone mode')
                sys.exit()
            load_channel_filter_from_db(peer_ip=peer, mongo_api=mongo_connection)

    # Starting api server
    if sys.version_info[0] == 2:
        from twisted.web.wsgi import WSGIResource
        LOG.info("Prepare RESTAPI service")
        resource = WSGIResource(reactor, reactor.getThreadPool(), app)
        site = Site(resource)
        try:
            reactor.listenTCP(CONF.rest.bind_port, site, interface=CONF.rest.bind_host)
            LOG.info("serving RESTAPI on http://%s:%s", CONF.rest.bind_host, CONF.rest.bind_port)
        except Exception as e:
            LOG.error(e, exc_info=True)
            sys.exit()

    for peer in all_peers:
        LOG.info('start peer, peer address=%s', peer)
        all_peers[peer].automatic_start()
    reactor.run()


def prepare_service(args=None):
    try:
        CONF(args=args, project='yabgp', version=version,
             default_config_files=['/etc/yabgp/yabgp.ini'])
    except cfg.ConfigFilesNotFoundError:
        CONF(args=args, project='yabgp', version=version)

    check_running_mode()
    log.init_log()
    LOG.info('Log (Re)opened.')
    LOG.info("Configuration:")
    cfg.CONF.log_opt_values(LOG, logging.INFO)
    try:
        get_bgp_config()
        check_msg_config()
    except Exception as e:
        LOG.error(e)
        LOG.debug(traceback.format_exc())
        sys.exit()

    LOG.info('Starting server in PID %s' % os.getpid())

    # write pid file
    if CONF.pid_file:
        with open(CONF.pid_file, 'w') as pid_file:
            pid_file.write(str(os.getpid()))
            LOG.info('create pid file: %s' % CONF.pid_file)
    prepare_twisted_service()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Set up the API server application instance
"""

import flask

from yabgp.api import v1
from yabgp.api import config
from oslo_config import cfg

app = flask.Flask('yabgp.api')
app.config['SECRET_KEY'] = 'cisco123'
app.register_blueprint(v1.blueprint, url_prefix='/v1')

cfg.CONF.register_cli_opts(config.rest_server_ops, group='rest')


@app.route('/')
def index():
    """API Root. Get the API version information.

    **Example request**:

    .. sourcecode:: http

      GET / HTTP/1.1
      Host: example.com
      Accept: application/json, text/javascript

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Vary: Accept
      Content-Type: text/json
      {
        "versions": {
        "values": [
            {
                "id": "v1",
                "links": "http://10.75.44.11:8801//v1",
                "status": "stable"
            }
        ]
       }
     }

    :status 200: the api can work.
    """
    base_url = flask.request.base_url
    available = [
        {'tag': 'v1'}]
    collected = [version_descriptor(base_url, v['tag']) for v in available]
    versions = {'versions': {'values': collected}}
    return flask.jsonify(versions)


def version_descriptor(base_url, version):
    url = version_url(base_url, version)
    return {
        'id': version,
        'links': url,
        'status': 'stable'
    }


def version_url(base_url, version_number):
    return '%s/%s' % (base_url, version_number)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" twisted message queue protocol
reference from https://github.com/pika/pika/blob/master/examples/twisted_service.py
"""

import logging

from pika import spec
from pika.adapters import twisted_connection
from twisted.internet.defer import inlineCallbacks

PREFETCH_COUNT = 2
LOG = logging.getLogger(__name__)


class PikaProtocol(twisted_connection.TwistedProtocolConnection):
    is_connected = False
    name = 'AMQP:Protocol'
    factory = None

    @inlineCallbacks
    def connected(self, connection):

        self.channel = yield connection.channel()
        yield self.channel.basic_qos(prefetch_count=PREFETCH_COUNT)
        self.is_connected = True
        for routing_key in self.factory.peer_list:
            LOG.info('routing key %s', routing_key)
            yield self.setup('', routing_key)

        # try to send all waiting messages
        LOG.info('connected, try to send all waiting message')
        self.send()

    @inlineCallbacks
    def setup(self, exchange, routing_key):
        """This function does the work to read from an exchange."""
        LOG.info("setup rabbitmq exchange or queue")
        if not exchange == '':
            yield self.channel.exchange_declare(exchange=exchange, type='topic', durable=True, auto_delete=False)
        else:
            yield self.channel.queue_declare(queue=routing_key, durable=True, auto_delete=False)

    def send(self):
        """If connected, send all waiting messages."""
        if self.is_connected:
            while len(self.factory.queued_messages) > 0:
                (exchange, r_key, message,) = self.factory.queued_messages.pop(0)
                self.send_message(exchange, r_key, message)

    @inlineCallbacks
    def send_message(self, exchange, routing_key, msg):
        """Send a single message."""
        LOG.debug('%s (%s): %s', exchange, routing_key, repr(msg))
        # yield self.channel.exchange_declare(exchange=exchange, type='topic', durable=True, auto_delete=False)
        prop = spec.BasicProperties(delivery_mode=2)
        try:
            yield self.channel.basic_publish(exchange=exchange, routing_key=routing_key, body=msg, properties=prop)
        except Exception as error:
            LOG.error('Error while sending message: %s', error)

    def connectionLost(self, reason):

        """Called when the associated connection was lost.

        :param reason: the reason of lost connection.
        """
        LOG.debug('Called connectionLost')
        self.is_connected = False
        LOG.info("Connection lost:%s", reason.getErrorMessage())



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Attribute MP_UNREACH_NLRI
"""

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeFlag
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute.nlri.ipv4_flowspec import IPv4FlowSpec
from yabgp.common import afn
from yabgp.common import safn
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons


class MpUnReachNLRI(Attribute):

    """
        This is an optional non-transitive attribute that can be used for the
    purpose of withdrawing multiple unfeasible routes from service.
        An UPDATE message that contains the MP_UNREACH_NLRI is not required
    to carry any other path attributes.

        MP_UNREACH_NLRI coding information
        +---------------------------------------------------------+
        | Address Family Identifier (2 octets)                    |
        +---------------------------------------------------------+
        | Subsequent Address Family Identifier (1 octet)          |
        +---------------------------------------------------------+
        | Withdrawn Routes (variable)                             |
        +---------------------------------------------------------+
    """

    ID = AttributeID.MP_UNREACH_NLRI
    FLAG = AttributeFlag.OPTIONAL

    @classmethod
    def parse(cls, value):
        try:
            afi, safi = struct.unpack('!HB', value[0:3])
        except Exception:
            raise excep.UpdateMessageError(sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                                           data='')

        nlri = value[3:]

        # for IPv4
        if afi == afn.AFNUM_INET:
            # BGP flow spec
            if safi == safn.SAFNUM_FSPEC_RULE:
                return IPv4FlowSpec().parse(value=nlri)
            else:
                return {'afi_safi': (afn.AFNUM_INET, safi),
                        'withdraw': repr(nlri)}

        return dict(afi_safi=(afi, safi), withdraw=nlri)

    @classmethod
    def construct(cls, value):

        """Construct a attribute

        :param value: python dictionary
        {'afi_safi': (1,128),
         'withdraw': []
        """
        afi, safi = value['afi_safi']
        if afi == afn.AFNUM_INET:
            if safi == safn.SAFNUM_LAB_VPNUNICAST:  # MPLS VPN
                pass
            elif safi == safn.SAFNUM_FSPEC_RULE:
                try:
                    nlri = IPv4FlowSpec().construct(value=value['withdraw'])
                    if nlri:
                        attr_value = struct.pack('!H', afi) + struct.pack('!B', safi) + \
                            nlri
                        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
                            + struct.pack('!B', len(attr_value)) + attr_value
                    else:
                        return None
                    pass
                except Exception:
                    raise excep.ConstructAttributeFailed(
                        reason='failed to construct attributes',
                        data=value
                    )
            else:
                raise excep.ConstructAttributeFailed(
                    reason='unsupport this sub address family',
                    data=value)
        else:
            raise excep.ConstructAttributeFailed(
                reason='unsupport this sub address family',
                data=value)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Origin attribute"""

import unittest

from yabgp.common.exception import UpdateMessageError
from yabgp.common.constants import ERR_MSG_UPDATE_INVALID_ORIGIN
from yabgp.message.attribute.origin import Origin


class TestOrigin(unittest.TestCase):

    def test_origin_igp(self):
        self.assertEqual(Origin.IGP, Origin.parse(value=b'\x00'))

    def test_origin_egp(self):
        self.assertEqual(Origin.EGP, Origin.parse(value=b'\x01'))

    def test_origin_incomplete(self):
        self.assertEqual(Origin.INCOMPLETE, Origin.parse(value=b'\x02'))

    def test_invalid_origin(self):

        self.assertRaises(UpdateMessageError, Origin.parse, b'\x03')
        try:
            Origin.parse(b'\x03')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_INVALID_ORIGIN)

    def test_construct(self):

        self.assertEqual(b'\x40\x01\x01\x00', Origin.construct(value=0))
        self.assertEqual(b'\x40\x01\x01\x00', Origin.construct(value=0))
        self.assertEqual(b'\x40\x01\x01\x01', Origin.construct(value=1))
        self.assertEqual(b'\x40\x01\x01\x02', Origin.construct(value=2))

        self.assertRaises(UpdateMessageError, Origin.construct, 3)
        try:
            Origin.construct(3)
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_INVALID_ORIGIN)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" basic config """

import logging
import sys
import os

from oslo_config import cfg

CONF = cfg.CONF

CONF.register_cli_opts([
    cfg.BoolOpt('standalone', default=True, help='The BGP Agent running mode'),
    cfg.StrOpt('pid-file', default=None, help='pid file name')
])

msg_process_opts = [
    cfg.BoolOpt('write_disk',
                default=True,
                help='Whether the BGP message is written to disk'),
    cfg.StrOpt('write_dir',
               default=os.path.join(os.environ['HOME'], 'data/bgp/'),
               help='The BGP messages storage path'),
    cfg.IntOpt('write_msg_max_size',
               default=500,
               help='The Max size of one BGP message file, the unit is MB'),
    cfg.BoolOpt('write_keepalive',
                default=False,
                help='Whether write keepalive message to disk'),
    cfg.StrOpt('format',
               default='json',
               choices=['json', 'list'],
               help='The output format of bgp messagees.')
]

CONF.register_opts(msg_process_opts, group='message')

bgp_config_opts = [
    cfg.IntOpt('peer_start_interval',
               default=10,
               help='The interval to start each BGP peer'),
    cfg.ListOpt('afi_safi',
                default=['ipv4'],
                help='The Global config for address family and sub address family'),
    cfg.BoolOpt('four_bytes_as',
                default=True,
                help='If support 4bytes AS'),
    cfg.BoolOpt('route_refresh',
                default=True,
                help='If support sending and receiving route refresh message'),
    cfg.BoolOpt('cisco_route_refresh',
                default=True,
                help='If support sending and receiving cisco route refresh message'),
    cfg.BoolOpt('enhanced_route_refresh',
                default=True,
                help='If support enhanced route refresh'),
    cfg.StrOpt('add_path',
               choices=['ipv4_send', 'ipv4_receive', 'ipv4_both'],
               help='BGP additional path feature and supported address family'),
    cfg.BoolOpt('graceful_restart',
                default=True,
                help='if support graceful restart'),
    cfg.BoolOpt('cisco_multi_session',
                default=True,
                help='if support cisco multi session'),
    cfg.DictOpt('running_config',
                default={},
                help='The running configuration for BGP'),
    cfg.StrOpt('config_file',
               help='BGP peers configuration file')
]

CONF.register_opts(bgp_config_opts, group='bgp')

bgp_peer_conf_cli_opts = [
    cfg.IntOpt('remote_as',
               help='The remote BGP peer AS number'),
    cfg.IntOpt('local_as',
               help='The Local BGP AS number'),
    cfg.StrOpt('remote_addr',
               help='The remote address of the peer'),
    cfg.StrOpt('local_addr',
               default='0.0.0.0',
               help='The local address of the BGP'),
    cfg.StrOpt('md5',
               help='The MD5 string use to auth',
               secret=True),
    cfg.BoolOpt('rib',
                default=False,
                help='Whether maintain BGP rib table'),
    cfg.StrOpt('tag',
               choices=['SRC', 'DST', 'BOTH'],
               help='The agent role tag'
               )
]

CONF.register_cli_opts(bgp_peer_conf_cli_opts, group='bgp')

LOG = logging.getLogger(__name__)


def get_bgp_config():
    """
    Get BGP running config
    :return:
    """
    # check bgp_conf_file
    if CONF.bgp.config_file:
        LOG.info('Try to load BGP configuration from %s', CONF.bgp.config_file)
        LOG.error('Failed to load BGP configuration')
        # TODO parse xml config file to get multi bgp config
        # will be supported in future
        sys.exit()
    else:
        # check bgp configuration from CLI input
        LOG.info('Try to load BGP configuration from CLI input')
        if CONF.bgp.local_as and CONF.bgp.remote_as and CONF.bgp.local_addr and CONF.bgp.remote_addr:
            CONF.bgp.running_config[CONF.bgp.remote_addr] = {
                'remote_as': CONF.bgp.remote_as,
                'remote_addr': CONF.bgp.remote_addr,
                'local_as': CONF.bgp.local_as,
                'local_addr': CONF.bgp.local_addr,
                'md5': CONF.bgp.md5,
                'afi_safi': CONF.bgp.afi_safi,
                'capability': {
                    'local': {
                        'four_bytes_as': CONF.bgp.four_bytes_as,
                        'route_refresh': CONF.bgp.route_refresh,
                        'cisco_route_refresh': CONF.bgp.cisco_route_refresh,
                        'enhanced_route_refresh': CONF.bgp.enhanced_route_refresh,
                        'graceful_restart': CONF.bgp.graceful_restart,
                        'cisco_multi_session': CONF.bgp.cisco_multi_session,
                        'add_path': CONF.bgp.add_path},
                    'remote': {}
                },
                'tag': CONF.bgp.tag
            }

            LOG.info('Get BGP running configuration for peer %s', CONF.bgp.remote_addr)
            for item in CONF.bgp.running_config[CONF.bgp.remote_addr]:
                if item == 'capability':
                    LOG.info('capability local:')
                    for capa in CONF.bgp.running_config[CONF.bgp.remote_addr][item]['local']:
                        LOG.info('-- %s: %s' % (
                            capa,
                            CONF.bgp.running_config[CONF.bgp.remote_addr][item]['local'][capa]
                        ))
                    continue

                LOG.info("%s = %s", item, CONF.bgp.running_config[CONF.bgp.remote_addr][item])
            return
        else:
            LOG.error('Please provide enough parameters!')
            sys.exit()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test MED attribute
"""

import unittest

from yabgp.common.exception import UpdateMessageError
from yabgp.common.constants import ERR_MSG_UPDATE_ATTR_LEN
from yabgp.message.attribute.med import MED


class TestMED(unittest.TestCase):

    def test_parse(self):

        med = MED.parse(value=b'\x00\x00\x00\xa0')
        self.assertEqual(med, 160)

        # invalid length
        self.assertRaises(UpdateMessageError, MED.parse,
                          b'\x0a\x0a\x0a\x01\x01')

    def test_parse_exception_attr_len(self):
        try:
            MED.parse(b'\x0a\x0a\x0a\x01\x01')
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_ATTR_LEN)

    def test_construct(self):

        med = MED.construct(value=100)
        self.assertEqual(med, b'\x80\x04\x04\x00\x00\x00d')

    def test_construct_exception_attr_len(self):

        # invalid med value
        self.assertRaises(UpdateMessageError, MED.construct, 4294967296)
        try:
            MED.construct(value=4294967296)
        except UpdateMessageError as e:
            self.assertEqual(e.sub_error, ERR_MSG_UPDATE_ATTR_LEN)

if __name__ == '__main__':
    unittest.main()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import logging
from functools import wraps

from oslo_config import cfg
from flask import request
import flask

from yabgp.common import constants as common_cons

LOG = logging.getLogger(__name__)


def log_request(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        LOG.info('API request url %s', request.url)
        if request.query_string:
            LOG.info('API query string %s', request.query_string)
        LOG.info('API request method %s', request.method)
        if request.method == 'POST':
            LOG.info('API POST data %s', request.json)
        LOG.debug('API request environ %s', request.environ)
        return f(*args, **kwargs)
    return decorated_function


def makesure_peer_establish(f):
    @wraps(f)
    def decorator(*args, **kwargs):
        value = kwargs['peer_ip']
        if _ready_to_send_msg(peer_ip=value):
            return f(*args, **kwargs)
        else:
            return flask.jsonify({
                'status': False,
                'code': "Please check the peer's state"
            })
    return decorator


def get_peer_conf_and_state(peer_ip=None):
    """
    get peer configuration and state
    :param peer_ip: peer ip address
    :return:
    """

    def get_peer_state(peer):
        one_peer_state = {key: cfg.CONF.bgp.running_config[peer][key] for key in [
            'remote_as', 'remote_addr', 'local_as', 'local_addr', 'capability']}
        fsm = cfg.CONF.bgp.running_config[peer]['factory'].fsm.state
        one_peer_state['fsm'] = common_cons.stateDescr[fsm]
        if fsm == common_cons.ST_ESTABLISHED:
            one_peer_state['uptime'] = time.time() - cfg.CONF.bgp.running_config[peer]['factory'].fsm.uptime
        else:
            one_peer_state['uptime'] = 0
        return one_peer_state

    if peer_ip:
        peer_state = {}
        if peer_ip in cfg.CONF.bgp.running_config:
            peer_state = get_peer_state(peer_ip)
        return {'peer': peer_state}
    # for multi peers
    result = {'peers': []}
    for peer_ip in cfg.CONF.bgp.running_config:
        result['peers'].append(get_peer_state(peer_ip))
    return result


def get_peer_msg_statistic(peer_ip=None):
    """
    get peer send and receive message statistic
    :param peer_ip: peer ip address
    :return:
    """

    if peer_ip and peer_ip in cfg.CONF.bgp.running_config:
        return {
            'send': cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.msg_sent_stat,
            'receive': cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.msg_recv_stat,
        }
    else:
        return {}


def _ready_to_send_msg(peer_ip):
    """
    check if the peer is ready to send message
    :param peer_ip:  peer ip address
    :return: if is ready, return is True, or False
    """
    if peer_ip and peer_ip in cfg.CONF.bgp.running_config:
        # check if the peer is established
        peer_state = get_peer_conf_and_state(peer_ip)
        if peer_state.get('peer').get('fsm') == common_cons.stateDescr[common_cons.ST_ESTABLISHED]:
            return True
    return False


def send_route_refresh(peer_ip, afi, safi, res):
    """
    send route refresh messages
    :param peer_ip: peer ip address
    :return: the sending results
    """
    try:
        if cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.send_route_refresh(
                afi=afi, safi=safi, res=res):
            return {
                'status': True
            }
    except Exception as e:
        LOG.error(e)
        return {
            'status': False,
            'code': 'failed when send this message out'
        }


def send_update(peer_ip, attr, nlri, withdraw):
    """
    send update message
    :param peer_ip: peer ip address
    :return:
    """
    # TODO check RIB out policy
    # TODO update RIB out table
    if cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.send_update({
            'attr': attr, 'nlri': nlri, 'withdraw': withdraw}):
        # TODO update ad rib out table
        # if nlri or withdraw:
        #     for prefix in nlri:
        #         if isinstance(prefix, basestring):
        #             cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.get_rib_out()['ipv4'][prefix] = attr

        return {
            'status': True
        }
    else:
        return {
            'status': False,
            'code': 'failed when send this message out'
        }


def get_adj_rib_in(peer_ip, afi_safi, prefix=None):
    rib_table = cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.get_rib_in().get(afi_safi)
    if prefix:
        attr = rib_table.get(prefix)
        if not attr:
            flask.abort(404)
        else:
            return attr
    return rib_table.keys()


def get_adj_rib_out(peer_ip, afi_safi, prefix=None):
    rib_table = cfg.CONF.bgp.running_config[peer_ip]['factory'].fsm.protocol.get_rib_out().get(afi_safi)
    if prefix:
        attr = rib_table.get(prefix)
        if not attr:
            flask.abort(404)
        else:
            return attr
    return rib_table.keys()



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""BGP Notification Message"""

import struct


class Notification(object):

    """
    notification
    """
    MSG_NOTIFICATION = 3

    @staticmethod
    def parse(message):
        """
        parse notification message

        :param message: input raw binary message data
        """
        error, suberror = struct.unpack('!BB', message[:2])
        data = message[2:]
        return error, suberror, data

    def construct_header(self, message):

        """Prepends the mandatory header to a constructed BGP message

        :param message:
        """
        #    16-octet     2-octet  1-octet
        # ---------------+--------+---------+------+
        #    Maker      | Length |  Type   |  msg |
        # ---------------+--------+---------+------+
        return b'\xff'*16 + struct.pack('!HB', len(message) + 19, self.MSG_NOTIFICATION) + message

    def construct(self, error, suberror=0, data=b''):

        """Constructs a BGP Notification message

        :param error:
        :param suberror:
        :param data:
        """

        msg = struct.pack('!BB', error, suberror) + data
        return self.construct_header(msg)



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import exception as excep
from yabgp.common import constants as bgp_cons


class Community(Attribute):
    """
        COMMUNITIES path attribute is an optional
    transitive attribute of variable length. The attribute consists of a
    set of four octet values, each of which specify a community. All
    routes with this attribute belong to the communities listed in the
    attribute.
        The COMMUNITIES attribute has Type Code 8.
        http://www.iana.org/assignments/bgp-well-known-communities/bgp-well-known-communities.xml
    """

    ID = AttributeID.COMMUNITY
    FLAG = AttributeFlag.OPTIONAL + AttributeFlag.TRANSITIVE

    @classmethod
    def parse(cls, value):
        """
        parse BGP community.
        :param value:
        """
        community = []
        if value:
            try:
                length = len(value) / 2
                value_list = list(struct.unpack('!%dH' % length, value))
                while value_list:
                    value_type = value_list[0] * 16 * 16 * 16 * 16 + value_list[1]
                    if value_type in bgp_cons.WELL_KNOW_COMMUNITY_INT_2_STR:
                        community.append(bgp_cons.WELL_KNOW_COMMUNITY_INT_2_STR[value_type])
                    else:
                        community.append("%s:%s" % (value_list[0], value_list[1]))
                    value_list = value_list[2:]
            except Exception:
                raise excep.UpdateMessageError(
                    sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                    data=value)
        return community

    @classmethod
    def construct(cls, value):
        """
        construct a COMMUNITY path attribute
        :param value:
        """
        community_hex = b''
        for community in value:
            if community.upper() in bgp_cons.WELL_KNOW_COMMUNITY_STR_2_INT:
                value = bgp_cons.WELL_KNOW_COMMUNITY_STR_2_INT[community.upper()]
                community_hex += struct.pack('!I', value)
            else:
                try:
                    value = community.split(':')
                    value = int(value[0]) * 16 * 16 * 16 * 16 + int(value[1])
                    community_hex += struct.pack('!I', value)
                except Exception:
                    raise excep.UpdateMessageError(
                        sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                        data=value
                    )

        return struct.pack('!B', cls.FLAG) + struct.pack('!B', cls.ID) \
            + struct.pack('!B', len(community_hex)) + community_hex



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" Test Update message"""

import unittest

from yabgp.common.constants import HDR_LEN
from yabgp.message.update import Update
from yabgp.common.exception import UpdateMessageError


class TestUpdate(unittest.TestCase):

    def test_parse_prefix_list(self):

        prefix_hex = b'\x13\xb8\x9d\xe0\x18E\xb3\xdd\x18E\xb3\xdc\x18\xd1f\xb2\x16Bpd\x18\xd06\xc2'
        nlri = ['184.157.224.0/19', '69.179.221.0/24', '69.179.220.0/24',
                '209.102.178.0/24', '66.112.100.0/22', '208.54.194.0/24']
        self.assertEqual(nlri, Update.parse_prefix_list(prefix_hex))

    def test_parse_prefix_list_with_addpath(self):
        prefix_hex = b'\x00\x00\x00\x01\x20\x05\x05\x05\x05\x00\x00\x00\x01\x20\xc0\xa8\x01\x05'
        nlri = [
            {'prefix': '5.5.5.5/32', 'path_id': 1},
            {'prefix': '192.168.1.5/32', 'path_id': 1}
        ]
        self.assertEqual(nlri, Update.parse_prefix_list(prefix_hex, True))

    def test_parse_prefix_mask_larger_than_32(self):
        prefix_hex = b'\x21\xb8\x9d\xe0\x18E\xb3\xdd\x18E\xb3\xdc\x18\xd1f\xb2\x16Bpd\x18\xd06\xc2'
        self.assertRaises(UpdateMessageError, Update.parse_prefix_list, prefix_hex)

    def test_construct_prefix_v4(self):
        nlri = ['184.157.224.1/32', '32.65.243.12/30', '89.232.254.0/23', '69.179.221.0/24',
                '61.172.0.0/16', '202.223.128.0/17', '156.152.0.0/15', '15.0.0.0/8',
                '209.102.178.0/24', '66.112.100.0/22', '208.54.194.0/24']
        nlri_hex = Update.construct_prefix_v4(nlri)
        self.assertEqual(nlri, Update.parse_prefix_list(nlri_hex))

    def test_construct_prefix_v4_addpath(self):
        nlri = [{'path_id': 1, 'prefix': '99.99.99.99/32'}]
        nlri_hex = Update.construct_prefix_v4(nlri, add_path=True)
        self.assertEqual(b'\x00\x00\x00\x01\x20\x63\x63\x63\x63', nlri_hex)

    def test_parse_attributes_ipv4(self):

        attr_hex = b'@\x01\x01\x00@\x02\x08\x02\x03\x00\x01\x00\x02\x00\x03@\x03\x04\xac\x10\x01\x0e\x80\x04\x04' \
                   b'\x00\x00\x00\x00@\x05\x04\x00\x00\x00d\x80\t\x04\xac\x10\x01\x0e\x80\n\x08\x02\x02\x02\x02dddd'
        attributes = {1: 0,
                      2: [(2, [1, 2, 3])],
                      3: '172.16.1.14',
                      4: 0,
                      5: 100,
                      9: '172.16.1.14',
                      10: ['2.2.2.2', '100.100.100.100']}
        self.assertEqual(attributes, Update.parse_attributes(attr_hex, False))
        self.assertRaises(UpdateMessageError, Update.parse_attributes, attr_hex, True)

    def test_construct_attributes_ipv4(self):
        attr = {
            1: 2,
            2: [(2, [701, 71])],
            3: '219.158.1.204',
            5: 100,
            6: b'',
            7: (71, '16.96.243.103'),
            8: ['4837:701', '4837:2100'],
            9: '219.158.1.204',
            10: ['219.158.1.209', '0.0.0.30']}
        attr_hex = Update.construct_attributes(attr, asn4=True)
        self.assertEqual(attr, Update.parse_attributes(attr_hex, asn4=True))

    def test_parse_and_construct_ipv4_unicast_2byteas(self):
        # 2 bytes asn
        msg_hex = b'\x00\x00\x00\x28\x40\x01\x01\x02\x40\x02\x0a\x02\x01\x00\x1e\x01\x02\x00\x0a\x00\x14\x40\x03\x04' \
                  b'\x0a\x00\x00\x09\x80\x04\x04\x00\x00\x00\x00\xc0\x07\x06\x00\x1e\x0a\x00\x00\x09\x15\xac\x10\x00'
        update = Update.parse(None, msg_hex)
        attributes = {1: 2, 2: [(2, [30]), (1, [10, 20])], 3: '10.0.0.9', 4: 0, 7: (30, '10.0.0.9')}
        self.assertEqual(attributes, update['attr'])
        self.assertEqual([], update['withdraw'])
        self.assertEqual(['172.16.0.0/21'], update['nlri'])
        self.assertEqual(msg_hex, Update.construct(msg_dict=update, asn4=False)[HDR_LEN:])

    def test_parse_ipv4_4byteas(self):
        # 4 bytes asn
        msg_hex = b'\x00\x00\x00\x30\x40\x01\x01\x02\x40\x02\x10\x02\x01\x00\x00\x00\x1e\x01\x02\x00\x00\x00\x0a\x00' \
                  b'\x00\x00\x14\x40\x03\x04\x0a\x00\x00\x09\x80\x04\x04\x00\x00\x00\x00\xc0\x07\x08\x00\x00\x00' \
                  b'\x1e\x0a\x00\x00\x09\x15\xac\x10\x00'
        update = Update.parse(None, msg_hex, True)
        attributes = {1: 2, 2: [(2, [30]), (1, [10, 20])], 3: '10.0.0.9', 4: 0, 7: (30, '10.0.0.9')}
        self.assertEqual(attributes, update['attr'])
        self.assertEqual([], update['withdraw'])
        self.assertEqual(['172.16.0.0/21'], update['nlri'])

    def test_parse_ipv4_addpath_update(self):
        msg_hex = b'\x00\x00\x00\x30\x40\x01\x01\x00\x40\x02\x06\x02\x01\x00\x00\xfb\xff\x40\x03\x04\x0a\x00\x0e' \
                  b'\x01\x80\x04\x04\x00\x00\x00\x00\x40\x05\x04\x00\x00\x00\x64\x80\x0a\x04\x0a\x00\x22' \
                  b'\x04\x80\x09\x04\x0a\x00\x0f\x01\x00\x00\x00\x01\x20\x05\x05\x05\x05\x00\x00\x00\x01' \
                  b'\x20\xc0\xa8\x01\x05'
        update = Update.parse(None, msg_hex, True, True)
        attributes = {1: 0, 2: [(2, [64511])], 3: '10.0.14.1', 4: 0, 5: 100, 9: '10.0.15.1', 10: ['10.0.34.4']}
        self.assertEqual(attributes, update['attr'])
        self.assertEqual([], update['withdraw'])
        self.assertEqual([
            {'path_id': 1, 'prefix': '5.5.5.5/32'},
            {'path_id': 1, 'prefix': '192.168.1.5/32'}], update['nlri'])

    def test_parse_ipv4_addpath_withdraw(self):
        msg_hex = b'\x00\x09\x00\x00\x00\x01\x20\x63\x63\x63\x63\x00\x00'
        update = Update.parse(None, msg_hex, True, True)
        self.assertEqual([{'path_id': 1, 'prefix': '99.99.99.99/32'}], update['withdraw'])


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2007 by Mark Bergsma <mark@nedworks.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" BGP Timer"""

# Twisted modules
from twisted.internet import reactor, error


class BGPTimer(object):
    """
    Timer class with a slightly different Timer interface than the
    Twisted DelayedCall interface
    """

    def __init__(self, call_able):

        self.delayed_call = None
        self.callable = call_able

    def cancel(self):

        """Cancels the timer if it was running, does nothing otherwise"""

        try:
            self.delayed_call.cancel()
        except (AttributeError, error.AlreadyCalled, error.AlreadyCancelled):
            pass

    def reset(self, seconds_fromnow):
        """Resets an already running timer, or starts it if it wasn't running.

        :param seconds_fromnow : restart timer
        """

        try:
            self.delayed_call.reset(seconds_fromnow)
        except (AttributeError, error.AlreadyCalled, error.AlreadyCancelled):
            self.delayed_call = reactor.callLater(seconds_fromnow, self.callable)

    def active(self):
        """Returns True if the timer was running, False otherwise."""

        try:
            return self.delayed_call.active()
        except AttributeError:
            return False



# Copyright 2015 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import struct

import netaddr

from yabgp.message.attribute import Attribute
from yabgp.message.attribute import AttributeID
from yabgp.message.attribute import AttributeFlag
from yabgp.common import constants as bgp_cons
from yabgp.common import exception as excep


class ClusterList(Attribute):
    """
        CLUSTER_LIST is a new, optional, non-transitive BGP attribute of Type
    code 10. It is a sequence of CLUSTER_ID values representing the
    reflection path that the route has passed.
    When an RR reflects a route, it MUST prepend the local CLUSTER_ID to
    the CLUSTER_LIST. If the CLUSTER_LIST is empty, it MUST create a new
    one. Using this attribute an RR can identify if the routing
    information has looped back to the same cluster due to
    misconfiguration. If the local CLUSTER_ID is found in the
    CLUSTER_LIST, the advertisement received SHOULD be ignored.
    (RFC 4456 Page 7)
    """

    ID = AttributeID.CLUSTER_LIST
    FLAG = AttributeFlag.OPTIONAL
    MULTIPLE = False

    @classmethod
    def parse(cls, value):

        """
        Parse culster list
        :param value
        """
        cluster_list = []
        if len(value) % 4 != 0:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=repr(value))
        while value:
            cluster_list.append(str(netaddr.IPAddress(struct.unpack('!I', value[0:4])[0])))
            value = value[4:]
        return cluster_list

    @classmethod
    def construct(cls, value):

        """
        construct a CLUSTER_LIST path attribute
        :param value:
        """
        cluster_raw = b''
        try:
            for cluster in value:
                cluster_raw += netaddr.IPAddress(cluster).packed
            return struct.pack("!B", cls.FLAG) + struct.pack('!B', cls.ID) \
                + struct.pack("!B", len(cluster_raw)) + cluster_raw
        except Exception:
            raise excep.UpdateMessageError(
                sub_error=bgp_cons.ERR_MSG_UPDATE_ATTR_LEN,
                data=struct.pack('B', cls.FLAG))
